import contextlib
import itertools
from collections import defaultdict, namedtuple

from django.core.checks import Error, Warning
from django.core.exceptions import FieldDoesNotExist
from django.db import NotSupportedError
from django.db.backends.utils import names_digest, split_identifier
from django.db.models import FloatField, Index, IntegerField
from django.db.models.lookups import BuiltinLookup
from django.db.models.sql.query import Query
from django.db.models.sql.where import AND, XOR, WhereNode
from pymongo import ASCENDING, DESCENDING
from pymongo.operations import IndexModel, SearchIndexModel

from django_mongodb_backend.fields import ArrayField

from .query_utils import process_rhs

MONGO_INDEX_OPERATORS = {
    "exact": "$eq",
    "gt": "$gt",
    "gte": "$gte",
    "lt": "$lt",
    "lte": "$lte",
    "in": "$in",
}

FieldColumn = namedtuple("FieldColumn", ["field", "column"])


def _get_condition_mql(self, model, schema_editor):
    """Analogous to Index._get_condition_sql()."""
    query = Query(model=model, alias_cols=False)
    where = query.build_where(self.condition)
    compiler = query.get_compiler(connection=schema_editor.connection)
    return where.as_mql_idx(compiler, schema_editor.connection)


def builtin_lookup_idx(self, compiler, connection):
    lhs_mql = self.lhs.target.column
    value = process_rhs(self, compiler, connection)
    try:
        operator = MONGO_INDEX_OPERATORS[self.lookup_name]
    except KeyError:
        raise NotSupportedError(
            f"MongoDB does not support the '{self.lookup_name}' lookup in indexes."
        ) from None
    return {lhs_mql: {operator: value}}


def get_field(model, field_name):
    """
    A version of Model_.meta.get_field() that can retrieve embedded model
    fields.
    """
    path = []
    base_model = model
    *parents, leaf = field_name.split(".")
    for i, name in enumerate(parents):
        field = model._meta.get_field(name)
        path.append(field.column)
        # For EmbeddedModelFields, advance to the embedded model and continue
        # to loop, searching for the next field.
        if hasattr(field, "embedded_model"):
            model = field.embedded_model
        # For PolymorphicEmbeddedModelFields, recurse into each embedded model
        # until the field is found.
        elif models := getattr(field, "embedded_models", None):
            for submodel in models:
                with contextlib.suppress(FieldDoesNotExist):
                    subfield = get_field(submodel, ".".join([*parents[i + 1 :], leaf]))
                    path.extend(subfield.column.split("."))
                    return FieldColumn(subfield.field, ".".join(path))
            raise FieldDoesNotExist(
                f"The models of field '{'.'.join(parents)}' have no field named '{leaf}'."
            )
        else:
            raise FieldDoesNotExist(f"{base_model.__name__} has no field named '{field_name}'.")
    # Add the final field.
    field = model._meta.get_field(leaf)
    path.append(field.column)
    return FieldColumn(field, ".".join(path))


def get_pymongo_index_model(self, model, schema_editor, field=None, column_prefix=""):
    """Return a pymongo IndexModel for this Django Index."""
    if self.contains_expressions:
        return None
    kwargs = {}
    filter_expression = defaultdict(dict)
    if self.condition:
        filter_expression.update(self._get_condition_mql(model, schema_editor))
    if filter_expression:
        kwargs["partialFilterExpression"] = filter_expression
    index_orders = (
        [(column_prefix + field.column, ASCENDING)]
        if field
        else [
            # order is "" if ASCENDING or "DESC" if DESCENDING (see
            # django.db.models.indexes.Index.fields_orders).
            (
                column_prefix + get_field(model, field_name).column,
                ASCENDING if order == "" else DESCENDING,
            )
            for field_name, order in self.fields_orders
        ]
    )
    return IndexModel(index_orders, name=self.name, **kwargs)


def where_node_idx(self, compiler, connection):
    if self.connector == AND:
        operator = "$and"
    elif self.connector == XOR:
        raise NotSupportedError("MongoDB does not support the '^' operator lookup in indexes.")
    else:
        operator = "$or"
    if self.negated:
        raise NotSupportedError("MongoDB does not support the '~' operator in indexes.")
    children_mql = []
    for child in self.children:
        mql = child.as_mql_idx(compiler, connection)
        children_mql.append(mql)
    if len(children_mql) == 1:
        mql = children_mql[0]
    elif len(children_mql) > 1:
        mql = {operator: children_mql}
    else:
        mql = {}
    return mql


class EmbeddedFieldIndexMixin:
    def check(self, model, connection):
        # The parent check reports E012 (nonexistent fields) based on top-level
        # fields. These errors are discarded and the checks redone, accounting
        # for embedded model fields.
        errors = [e for e in super().check(model, connection) if e.id != "models.E012"]
        return errors + self._check_local_fields(model)

    def _check_local_fields(self, model):
        errors = []
        forward_fields = self._get_forward_fields(model)
        for field_name in self.fields:
            if field_name not in forward_fields:
                errors.append(
                    Error(
                        f"'{self.meta_option_name}' refers to the nonexistent "
                        f"field '{field_name}'.",
                        obj=model,
                        id="models.E012",
                    )
                )
        return errors

    def _get_forward_fields(self, model):
        """
        Return the set of forward field paths for the given model, including
        embedded fields.
        """
        forward_fields = set()
        for field in model._meta._get_fields(reverse=False):
            # Recurse into embedded models and flatten their forward fields
            # using dotted paths (e.g. "field.subfield").
            if embedded_model := getattr(field, "embedded_model", None):
                embedded_models = [embedded_model]
            elif embedded_models := getattr(field, "embedded_models", None):
                pass
            else:
                embedded_models = []
            for embedded_model in embedded_models:
                sub_forward_fields = self._get_forward_fields(embedded_model)
                for column in sub_forward_fields:
                    forward_fields.add(f"{field.name}.{column}")
                    if hasattr(field, "attname"):
                        forward_fields.add(f"{field.attname}.{column}")
            # For each field, both the public field name and its attribute name
            # (attname), when present, are included in order to mirror Django's
            # field resolution.
            forward_fields.add(field.name)
            if hasattr(field, "attname"):
                forward_fields.add(field.attname)
        return forward_fields


class EmbeddedFieldIndex(EmbeddedFieldIndexMixin, Index):
    meta_option_name = "indexes"

    def set_name_with_model(self, model):
        """
        Generate a unique name for the index.

        The name is divided into three parts: the table name, the field name,
        and a hash plus suffix. Each part is truncated to fit within Django's
        index name length constraints.

        This method overrides Django's base implementation which uses
        model._meta.get_field(field_name).column, which doesn't work for
        embedded fields.
        """
        _, table_name = split_identifier(model._meta.db_table)
        field_names_with_order = [
            (f"-{column_name}" if order else column_name)
            for column_name, order in self.fields_orders
        ]
        hash_data = [table_name, *field_names_with_order, self.suffix]
        self.name = (
            f"{table_name[:11]}_{self.fields_orders[0][0][:7]}_"
            f"{names_digest(*hash_data, length=6)}_{self.suffix}"
        )
        if self.name[0] == "_" or self.name[0].isdigit():
            self.name = f"D{self.name[1:]}"


class SearchIndex(Index):
    suffix = "six"

    def __init__(
        self, *, fields=(), field_mappings=None, name=None, analyzer=None, search_analyzer=None
    ):
        if field_mappings and not isinstance(field_mappings, dict):
            raise ValueError(
                "field_mappings must be a dictionary mapping field names to their "
                "MongoDB Search index options."
            )
        if analyzer and not isinstance(analyzer, str):
            raise ValueError(f"analyzer must be a string; got: {type(analyzer)}.")
        if search_analyzer and not isinstance(search_analyzer, str):
            raise ValueError(f"search_analyzer must be a string; got: {type(search_analyzer)}.")
        self.field_mappings = field_mappings
        self.analyzer = analyzer
        self.search_analyzer = search_analyzer
        if field_mappings:
            if fields:
                raise ValueError("Cannot provide fields and field_mappings.")
            fields = [*self.field_mappings.keys()]
        super().__init__(fields=fields, name=name)

    def deconstruct(self):
        path, args, kwargs = super().deconstruct()
        if self.field_mappings:
            kwargs["field_mappings"] = self.field_mappings
            del kwargs["fields"]
        if self.analyzer:
            kwargs["analyzer"] = self.analyzer
        if self.search_analyzer:
            kwargs["search_analyzer"] = self.search_analyzer
        return path, args, kwargs

    def check(self, model, connection):
        errors = []
        if not connection.features.supports_search:
            errors.append(
                Warning(
                    f"This MongoDB server does not support {self.__class__.__name__}.",
                    hint=(
                        "The index won't be created. Use a Search-enabled version of MongoDB, "
                        "or silence this warning if you don't care about it."
                    ),
                    obj=model,
                    id="mongodb.indexes.search.W001",
                )
            )
        return errors

    def search_index_data_types(self, db_type):
        """
        Map a model field's type to search index type.
        https://www.mongodb.com/docs/atlas/atlas-search/define-field-mappings/#data-types
        """
        if db_type in {"double", "int", "long"}:
            return "number"
        if db_type == "binData":
            return "string"
        if db_type == "bool":
            return "boolean"
        if db_type == "object":
            return "document"
        if db_type == "array":
            return "embeddedDocuments"
        return db_type

    def get_pymongo_index_model(self, model, schema_editor, field=None, column_prefix=""):
        if not schema_editor.connection.features.supports_search:
            return None
        fields = {}
        for field_name, _ in self.fields_orders:
            field_path = column_prefix + model._meta.get_field(field_name).column
            if self.field_mappings:
                fields[field_path] = self.field_mappings[field_name]
            else:
                field = model._meta.get_field(field_name)
                type_ = self.search_index_data_types(field.db_type(schema_editor.connection))
                fields[field_path] = {"type": type_}
        extra = {}
        if self.analyzer:
            extra["analyzer"] = self.analyzer
        if self.search_analyzer:
            extra["searchAnalyzer"] = self.search_analyzer
        return SearchIndexModel(
            definition={"mappings": {"dynamic": False, "fields": fields}, **extra},
            name=self.name,
        )


class VectorSearchIndex(SearchIndex):
    suffix = "vsi"
    VALID_FIELD_TYPES = frozenset(("boolean", "date", "number", "objectId", "string", "uuid"))
    VALID_SIMILARITIES = frozenset(("cosine", "dotProduct", "euclidean"))

    def __init__(self, *, fields=(), name=None, similarities):
        super().__init__(fields=fields, name=name)
        self.similarities = similarities
        self._multiple_similarities = isinstance(similarities, (tuple, list))
        for func in similarities if self._multiple_similarities else (similarities,):
            if func not in self.VALID_SIMILARITIES:
                raise ValueError(
                    f"'{func}' isn't a valid similarity function "
                    f"({', '.join(sorted(self.VALID_SIMILARITIES))})."
                )
        seen_fields = set()
        for field_name, _ in self.fields_orders:
            if field_name in seen_fields:
                raise ValueError(f"Field '{field_name}' is duplicated in fields.")
            seen_fields.add(field_name)

    def check(self, model, connection):
        errors = super().check(model, connection)
        num_arrayfields = 0
        for field_name, _ in self.fields_orders:
            field = model._meta.get_field(field_name)
            if isinstance(field, ArrayField):
                num_arrayfields += 1
                try:
                    int(field.size)
                except (ValueError, TypeError):
                    errors.append(
                        Error(
                            f"VectorSearchIndex requires 'size' on field '{field_name}'.",
                            obj=model,
                            id="mongodb.indexes.search.E002",
                        )
                    )
                if not isinstance(field.base_field, (FloatField, IntegerField)):
                    errors.append(
                        Error(
                            "VectorSearchIndex requires the base field of "
                            f"ArrayField '{field.name}' to be FloatField or "
                            "IntegerField but is "
                            f"{field.base_field.get_internal_type()}.",
                            obj=model,
                            id="mongodb.indexes.search.E003",
                        )
                    )
            else:
                search_type = self.search_index_data_types(field.db_type(connection))
                if search_type not in self.VALID_FIELD_TYPES:
                    errors.append(
                        Error(
                            "VectorSearchIndex does not support field "
                            f"'{field_name}' ({field.get_internal_type()}).",
                            obj=model,
                            id="mongodb.indexes.search.E004",
                            hint=f"Allowed types are {', '.join(sorted(self.VALID_FIELD_TYPES))}.",
                        )
                    )
        if self._multiple_similarities and num_arrayfields != len(self.similarities):
            errors.append(
                Error(
                    f"VectorSearchIndex requires the same number of similarities "
                    f"and vector fields; {model._meta.object_name} has "
                    f"{num_arrayfields} ArrayField(s) but similarities "
                    f"has {len(self.similarities)} element(s).",
                    obj=model,
                    id="mongodb.indexes.search.E005",
                )
            )
        if num_arrayfields == 0:
            errors.append(
                Error(
                    "VectorSearchIndex requires at least one ArrayField to store vector data.",
                    obj=model,
                    id="mongodb.indexes.search.E006",
                    hint="If you want to perform search operations without vectors, "
                    "use SearchIndex instead.",
                )
            )
        return errors

    def deconstruct(self):
        path, args, kwargs = super().deconstruct()
        kwargs["similarities"] = self.similarities
        return path, args, kwargs

    def get_pymongo_index_model(self, model, schema_editor, field=None, column_prefix=""):
        if not schema_editor.connection.features.supports_search:
            return None
        similarities = (
            itertools.cycle([self.similarities])
            if not self._multiple_similarities
            else iter(self.similarities)
        )
        fields = []
        for field_name, _ in self.fields_orders:
            field_ = model._meta.get_field(field_name)
            field_path = column_prefix + model._meta.get_field(field_name).column
            mappings = {"path": field_path}
            if isinstance(field_, ArrayField):
                mappings.update(
                    {
                        "type": "vector",
                        "numDimensions": int(field_.size),
                        "similarity": next(similarities),
                    }
                )
            else:
                mappings["type"] = "filter"
            fields.append(mappings)
        return SearchIndexModel(definition={"fields": fields}, name=self.name, type="vectorSearch")


def register_indexes():
    BuiltinLookup.as_mql_idx = builtin_lookup_idx
    Index._get_condition_mql = _get_condition_mql
    Index.get_pymongo_index_model = get_pymongo_index_model
    WhereNode.as_mql_idx = where_node_idx



import itertools
from collections import defaultdict

from bson import SON, ObjectId, json_util
from django.core.exceptions import EmptyResultSet, FieldError, FullResultSet
from django.db import IntegrityError, NotSupportedError
from django.db.models import Count
from django.db.models.aggregates import Aggregate, Variance
from django.db.models.expressions import Case, Col, OrderBy, Ref, Value, When
from django.db.models.functions.comparison import Coalesce
from django.db.models.functions.math import Power
from django.db.models.lookups import IsNull
from django.db.models.sql import compiler
from django.db.models.sql.constants import GET_ITERATOR_CHUNK_SIZE, MULTI, SINGLE
from django.db.models.sql.datastructures import BaseTable
from django.db.models.sql.where import AND, OR, XOR, ExtraWhere, NothingNode, WhereNode
from django.utils.functional import cached_property
from pymongo import ASCENDING, DESCENDING

from .expressions.search import SearchExpression, SearchVector
from .query import MongoQuery, wrap_database_errors
from .query_utils import is_constant_value, is_direct_value


class SQLCompiler(compiler.SQLCompiler):
    """Base class for all Mongo compilers."""

    query_class = MongoQuery
    PARENT_FIELD_TEMPLATE = "parent__field__{}"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.aggregation_pipeline = None
        # Map columns to their subquery indices.
        self.column_indices = {}
        # A list of OrderBy objects for this query.
        self.order_by_objs = None
        self.subqueries = []
        # Search stage.
        self.search_pipeline = []
        # Does the aggregation have no GROUP BY fields and need wrapping?
        self.needs_wrap_aggregation = False
        # The MQL equivalent to a SQL HAVING clause.
        self.having_match_mql = None

    def _get_group_alias_column(self, expr, annotation_group_idx):
        """Generate a dummy field for use in the ids fields in $group."""
        replacement = None
        if isinstance(expr, Col):
            col = expr
        else:
            # If the column is a composite expression, create a field for it.
            alias = f"__annotation_group{next(annotation_group_idx)}"
            col = self._get_column_from_expression(expr, alias)
            replacement = col
        return col.target.column, replacement

    def _get_column_from_expression(self, expr, alias):
        """
        Create a column named `alias` from the given expression to hold the
        aggregate value.
        """
        column_target = expr.output_field.clone()
        column_target.db_column = alias
        column_target.set_attributes_from_name(alias)
        return Col(self.collection_name, column_target)

    def _get_replace_expr(self, sub_expr, group, alias):
        column_target = sub_expr.output_field.clone()
        column_target.db_column = alias
        column_target.set_attributes_from_name(alias)
        inner_column = Col(self.collection_name, column_target)
        if getattr(sub_expr, "distinct", False):
            # If the expression should return distinct values, use $addToSet to
            # deduplicate.
            rhs = sub_expr.as_mql(
                self, self.connection, resolve_inner_expression=True, as_expr=True
            )
            group[alias] = {"$addToSet": rhs}
            replacing_expr = sub_expr.copy()
            replacing_expr.set_source_expressions([inner_column, None])
        else:
            group[alias] = sub_expr.as_mql(self, self.connection, as_expr=True)
            replacing_expr = inner_column
        # Count must return 0 rather than null.
        if isinstance(sub_expr, Count):
            replacing_expr = Coalesce(replacing_expr, 0)
        # Variance = StdDev^2
        if isinstance(sub_expr, Variance):
            replacing_expr = Power(replacing_expr, 2)
        return replacing_expr

    def _prepare_expressions_for_pipeline(self, expression, target, annotation_group_idx):
        """
        Prepare expressions for the aggregation pipeline.

        Handle the computation of aggregation functions used by various
        expressions. Separate and create intermediate columns, and replace
        nodes to simulate a group by operation.

        MongoDB's $group stage doesn't allow operations over the aggregator,
        e.g. COALESCE(AVG(field), 3). However, it supports operations inside
        the aggregation, e.g. AVG(number * 2).

        Handle the first case by splitting the computation into stages: compute
        the aggregation first, then apply additional operations in a subsequent
        stage by replacing the aggregate expressions with new columns prefixed
        by `__aggregation`.
        """
        replacements = {}
        group = {}
        for sub_expr in self._get_aggregate_expressions(expression):
            alias = (
                f"__aggregation{next(annotation_group_idx)}" if sub_expr != expression else target
            )
            replacements[sub_expr] = self._get_replace_expr(sub_expr, group, alias)
        return replacements, group

    def _prepare_search_expressions_for_pipeline(self, expression, search_idx, replacements):
        """
        Collect and prepare unique search expressions for inclusion in an
        aggregation pipeline.

        Iterate over all search sub-expressions of the given expression.
        Assigning a unique alias to each and map them to their replacement
        expressions.
        """
        searches = {}
        for sub_expr in self._get_search_expressions(expression):
            if sub_expr not in replacements:
                alias = f"__search_expr.search{next(search_idx)}"
                replacements[sub_expr] = self._get_replace_expr(sub_expr, searches, alias)

    def _prepare_search_query_for_aggregation_pipeline(self, order_by):
        """
        Prepare expressions for the search pipeline.

        Handle the computation of search functions used by various expressions.
        Separate and create intermediate columns, and replace nodes to simulate
        a search operation.

        To apply operations over the $search or $searchVector stages, compute
        the $search or $vectorSearch first, then apply additional operations in
        a subsequent stage by replacing the aggregate expressions with a new
        document field prefixed by `__search_expr.search#`.
        """
        replacements = {}
        annotation_group_idx = itertools.count(start=1)
        for expr in self.query.annotation_select.values():
            self._prepare_search_expressions_for_pipeline(expr, annotation_group_idx, replacements)
        for expr, _ in order_by:
            self._prepare_search_expressions_for_pipeline(expr, annotation_group_idx, replacements)
        self._prepare_search_expressions_for_pipeline(
            self.having, annotation_group_idx, replacements
        )
        self._prepare_search_expressions_for_pipeline(
            self.get_where(), annotation_group_idx, replacements
        )
        return replacements

    def _prepare_annotations_for_aggregation_pipeline(self, order_by):
        """Prepare annotations for the aggregation pipeline."""
        replacements = {}
        group = {}
        annotation_group_idx = itertools.count(start=1)
        for target, expr in self.query.annotation_select.items():
            if expr.contains_aggregate:
                new_replacements, expr_group = self._prepare_expressions_for_pipeline(
                    expr, target, annotation_group_idx
                )
                replacements.update(new_replacements)
                group.update(expr_group)
        for expr, _ in order_by:
            if expr.contains_aggregate:
                new_replacements, expr_group = self._prepare_expressions_for_pipeline(
                    expr, None, annotation_group_idx
                )
                replacements.update(new_replacements)
                group.update(expr_group)
        having_replacements, having_group = self._prepare_expressions_for_pipeline(
            self.having, None, annotation_group_idx
        )
        replacements.update(having_replacements)
        group.update(having_group)
        return group, replacements

    def _get_group_expressions(self, order_by):
        if self.query.group_by is None:
            return []
        seen = set()
        expressions = set()
        if self.query.group_by is not True:
            # If group_by isn't True, then it's a list of expressions.
            for expr in self.query.group_by:
                if not hasattr(expr, "as_sql"):
                    expr = self.query.resolve_ref(expr)
                if isinstance(expr, Ref):
                    if expr.refs not in seen:
                        seen.add(expr.refs)
                        expressions.add(expr.source)
                else:
                    expressions.add(expr)
        for expr, _, alias in self.select:
            # Skip members that are already grouped.
            if alias not in seen:
                expressions |= set(expr.get_group_by_cols())
        if not self._meta_ordering:
            for expr, (_, _, is_ref) in order_by:
                # Skip references.
                if not is_ref:
                    expressions |= set(expr.get_group_by_cols())
        having_group_by = self.having.get_group_by_cols() if self.having else ()
        for expr in having_group_by:
            expressions.add(expr)
        return expressions

    def _get_group_id_expressions(self, order_by):
        """Generate group ID expressions for the aggregation pipeline."""
        replacements = {}
        group_expressions = self._get_group_expressions(order_by)
        if not group_expressions:
            ids = None
        else:
            annotation_group_idx = itertools.count(start=1)
            ids = {}
            columns = []
            for col in group_expressions:
                alias, replacement = self._get_group_alias_column(col, annotation_group_idx)
                columns.append((alias, col))
                if replacement is not None:
                    replacements[col] = replacement
                    if isinstance(col, Ref):
                        replacements[col.source] = replacement
            ids = self.get_project_fields(tuple(columns), force_expression=True)
        return ids, replacements

    def _build_aggregation_pipeline(self, ids, group):
        """Build the aggregation pipeline for grouping."""
        pipeline = []
        if not ids:
            pipeline.append({"$group": {"_id": None, **group}})
            # The aggregation must be wrapped if there are no group by ids and
            # no having clause.
            self.needs_wrap_aggregation = not bool(self.having)
        else:
            group["_id"] = ids
            pipeline.append({"$group": group})
            projected_fields = {key: f"$_id.{key}" for key in ids}
            pipeline.append({"$addFields": projected_fields})
            if "_id" not in projected_fields:
                pipeline.append({"$unset": "_id"})
        return pipeline

    def _compound_searches_queries(self, search_replacements):
        """
        Build a query pipeline from a mapping of search expressions to result
        columns.

        Currently only a single $search or $vectorSearch expression is
        supported. Combining multiple search expressions raises ValueError.

        This method will eventually support hybrid search by allowing the
        combination of $search and $vectorSearch operations.
        """
        if not search_replacements:
            return []
        if len(search_replacements) > 1:
            has_search = any(not isinstance(search, SearchVector) for search in search_replacements)
            has_vector_search = any(
                isinstance(search, SearchVector) for search in search_replacements
            )
            if has_search and has_vector_search:
                raise ValueError(
                    "Cannot combine a `$vectorSearch` with a `$search` operator. "
                    "If you need to combine them, consider restructuring your query logic or "
                    "running them as separate queries."
                )
            if has_vector_search:
                raise ValueError(
                    "Cannot combine two `$vectorSearch` operator. "
                    "If you need to combine them, consider restructuring your query logic or "
                    "running them as separate queries."
                )
            raise ValueError(
                "Only one $search operation is allowed per query. "
                f"Received {len(search_replacements)} search expressions. "
                "To combine multiple search expressions, use either a CompoundExpression for "
                "fine-grained control or CombinedSearchExpression for simple logical combinations."
            )
        pipeline = []
        for search, result_col in search_replacements.items():
            score_function = (
                "vectorSearchScore" if isinstance(search, SearchVector) else "searchScore"
            )
            pipeline.extend(
                [
                    search.as_mql(self, self.connection),
                    {
                        "$addFields": {
                            result_col.as_mql(self, self.connection): {"$meta": score_function}
                        }
                    },
                ]
            )
        return pipeline

    def pre_sql_setup(self, with_col_aliases=False):
        extra_select, order_by, group_by = super().pre_sql_setup(with_col_aliases=with_col_aliases)
        search_replacements = self._prepare_search_query_for_aggregation_pipeline(order_by)
        group, group_replacements = self._prepare_annotations_for_aggregation_pipeline(order_by)
        all_replacements = {**search_replacements, **group_replacements}
        self.search_pipeline = self._compound_searches_queries(search_replacements)
        # query.group_by is either:
        # - None: no GROUP BY
        # - True: group by select fields
        # - a list of expressions to group by.
        if group or self.query.group_by:
            ids, replacements = self._get_group_id_expressions(order_by)
            all_replacements.update(replacements)
            pipeline = self._build_aggregation_pipeline(ids, group)
            if self.having:
                having = self.having.replace_expressions(all_replacements).as_mql(
                    self, self.connection
                )
                # Add HAVING subqueries.
                for query in self.subqueries or ():
                    pipeline.extend(query.get_pipeline())
                # Remove the added subqueries.
                self.subqueries = []
                pipeline.append({"$match": having})
            self.aggregation_pipeline = pipeline
        self.annotations = {
            target: expr.replace_expressions(all_replacements)
            for target, expr in self.query.annotation_select.items()
        }
        self.order_by_objs = [expr.replace_expressions(all_replacements) for expr, _ in order_by]
        if (where := self.get_where()) and search_replacements:
            self.set_where(where.replace_expressions(search_replacements))
        return extra_select, order_by, group_by

    def get_project_columns(self, columns):
        # In order to construct less verbose queries, avoid $project if it's
        # unneeded. The columns must be projected if...
        needs_projection = (
            # The query has annotations (which appear in $project)
            self.query.annotations
            # A subset of columns are selected (e.g. QuerySet.values())
            or not self.query.default_cols
            # It's a distinct query (e.g. QuerySet.distinct())
            or self.query.distinct
            # The query has a select mask (e.g. QuerySet.defer()/only())
            or self.query.get_select_mask()
        )
        return columns if needs_projection else None

    def execute_sql(
        self, result_type=MULTI, chunked_fetch=False, chunk_size=GET_ITERATOR_CHUNK_SIZE
    ):
        self.pre_sql_setup()
        try:
            query = self.build_query(self.get_project_columns(self.columns))
        except EmptyResultSet:
            return iter([]) if result_type == MULTI else None

        cursor = query.get_cursor()
        if result_type == SINGLE:
            try:
                obj = cursor.next()
            except StopIteration:
                return None  # No result
            else:
                return self._make_result(obj, self.columns)
        # result_type is MULTI
        result = self.cursor_iter(cursor, chunk_size, self.columns)
        if not chunked_fetch:
            # If using non-chunked reads, read data into memory.
            return list(result)
        return result

    def results_iter(
        self,
        results=None,
        tuple_expected=False,
        chunked_fetch=False,
        chunk_size=GET_ITERATOR_CHUNK_SIZE,
    ):
        """
        Return an iterator over the results from executing query given
        to this compiler. Called by QuerySet methods.

        This method is copied from the superclass with one modification: the
        `if tuple_expected` block is deindented so that the result of
        _make_result() (a list) is cast to tuple as needed. For SQL database
        drivers, tuple results come from cursor.fetchmany(), so the cast is
        only needed there when apply_converters() casts the tuple to a list.
        This customized method could be removed if _make_result() cast its
        return value to a tuple, but that would be more expensive since that
        cast is not always needed.
        """
        if results is None:
            # QuerySet.values() or values_list()
            results = self.execute_sql(MULTI, chunked_fetch=chunked_fetch, chunk_size=chunk_size)

        fields = [s[0] for s in self.select[0 : self.col_count]]
        converters = self.get_converters(fields)
        rows = itertools.chain.from_iterable(results)
        if converters:
            rows = self.apply_converters(rows, converters)
        if tuple_expected:
            rows = map(tuple, rows)
        return rows

    def _make_result(self, entity, columns):
        """
        Decode values for the given fields from the database entity.

        The entity is assumed to be a dict using field database column
        names as keys.
        """
        result = []
        for name, col in columns:
            column_alias = getattr(col, "alias", None)
            obj = (
                # Use the related object...
                entity.get(column_alias, {})
                # ...if this column refers to an object for select_related().
                if column_alias is not None and column_alias != self.collection_name
                else entity
            )
            result.append(obj.get(name))
        return result

    def cursor_iter(self, cursor, chunk_size, columns):
        """Yield chunks of results from cursor."""
        chunk = []
        for row in cursor:
            chunk.append(self._make_result(row, columns))
            if len(chunk) == chunk_size:
                yield chunk
                chunk = []
        yield chunk

    def check_query(self):
        """Check if the current query is supported by the database."""
        if self.query.extra:
            if any(key.startswith("_prefetch_related_") for key in self.query.extra):
                raise NotSupportedError("QuerySet.prefetch_related() is not supported on MongoDB.")
            raise NotSupportedError("QuerySet.extra() is not supported on MongoDB.")

    def build_query(self, columns=None):
        """Check if the query is supported and prepare a MongoQuery."""
        self.check_query()
        query = self.query_class(self)
        ordering_fields, sort_ordering, extra_fields = self._get_ordering()
        query.ordering = sort_ordering
        if self.query.combinator:
            if not getattr(self.connection.features, f"supports_select_{self.query.combinator}"):
                raise NotSupportedError(
                    f"{self.query.combinator} is not supported on this database backend."
                )
            query.combinator_pipeline = self.get_combinator_queries()
        else:
            if self.query.distinct:
                # If query is distinct, build a $group stage for distinct
                # fields, then set project fields based on the grouped _id.
                distinct_fields = self.get_project_fields(
                    columns, ordering_fields, force_expression=True
                )
                if not query.aggregation_pipeline:
                    query.aggregation_pipeline = []
                query.aggregation_pipeline.extend(
                    [
                        {"$group": {"_id": distinct_fields}},
                        {"$project": {key: f"$_id.{key}" for key in distinct_fields}},
                    ]
                )
            else:
                # Otherwise, project fields without grouping.
                query.project_fields = self.get_project_fields(columns, ordering_fields)
            # If columns is None, then get_project_fields() won't add
            # ordering_fields to $project. Use $addFields (extra_fields) instead.
            if columns is None:
                extra_fields += ordering_fields
            query.lookup_pipeline = self.get_lookup_pipeline()
            where = self.get_where()
            try:
                match_mql = where.as_mql(self, self.connection) if where else {}
            except FullResultSet:
                query.match_mql = {}
            else:
                query.match_mql = match_mql
        if extra_fields:
            query.extra_fields = self.get_project_fields(extra_fields, force_expression=True)
        query.subqueries = self.subqueries
        return query

    @cached_property
    def columns(self):
        """
        Return a tuple of (name, expression) with the columns and annotations
        which should be loaded by the query.
        """
        select_mask = self.query.get_select_mask()
        columns = (
            self.get_default_columns(select_mask) if self.query.default_cols else self.query.select
        )

        annotation_idx = 1

        def project_field(column):
            nonlocal annotation_idx
            if hasattr(column, "target"):
                # column is a Col.
                target = column.target.column
            # Handle Order By columns as refs columns.
            elif isinstance(column, OrderBy) and isinstance(column.expression, Ref):
                target = column.expression.refs
            else:
                # column is a Transform in values()/values_list() that needs a
                # name for $proj.
                target = f"__annotation{annotation_idx}"
                annotation_idx += 1
            return target, column

        selected = []
        if self.query.selected is None:
            selected = [
                *(project_field(col) for col in columns),
                *self.annotations.items(),
            ]
        else:
            for expression in self.query.selected.values():
                # Reference to an annotation.
                if isinstance(expression, str):
                    alias, expression = expression, self.annotations[expression]
                # Reference to a column.
                elif isinstance(expression, int):
                    alias, expression = project_field(columns[expression])
                selected.append((alias, expression))
        # Populate QuerySet.select_related() data.
        related_columns = []
        if self.query.select_related:
            self.get_related_selections(related_columns, select_mask)
            if related_columns:
                related_columns, _ = zip(*related_columns, strict=True)
        return tuple(selected) + tuple(map(project_field, related_columns))

    @cached_property
    def base_table(self):
        return next(
            (
                v
                for k, v in self.query.alias_map.items()
                if isinstance(v, BaseTable) and self.query.alias_refcount[k]
            ),
            None,
        )

    @cached_property
    def collection_name(self):
        if self.base_table:
            return self.base_table.table_alias or self.base_table.table_name
        # Use a dummy collection if the query doesn't specify a table.
        # For Constraint.validate():
        # SELECT 1 WHERE EXISTS(subquery checking if a constraint is violated)
        # is translated as:
        # [{"$facet": {"__null": []}},
        #  {"$lookup": {"the subquery"}},
        #  {"$match": {"condition to check from the subquery"}}]
        query = self.query_class(self)
        # The "__null" document is a placeholder so that the subquery has
        # somewhere to return its results.
        query.aggregation_pipeline = [{"$facet": {"__null": []}}]
        self.subqueries.insert(0, query)
        return "__null"

    @cached_property
    def collection(self):
        return self.connection.get_collection(self.collection_name)

    def get_combinator_queries(self):
        parts = []
        compilers = [
            query.get_compiler(self.using, self.connection, self.elide_empty)
            for query in self.query.combined_queries
        ]
        main_query_fields, _ = zip(*self.columns, strict=True)
        combinator = self.query.combinator
        for compiler_ in compilers:
            try:
                # If the columns list is limited, then all combined queries
                # must have the same columns list. Set the selects defined on
                # the query on all combined queries, if not already set.
                selected = self.query.selected
                if selected is not None and compiler_.query.selected is None:
                    compiler_.query = compiler_.query.clone()
                    compiler_.query.set_values(selected)
                elif not compiler_.query.values_select and self.query.values_select:
                    compiler_.query = compiler_.query.clone()
                    compiler_.query.set_values(
                        (
                            *self.query.extra_select,
                            *self.query.values_select,
                            *self.query.annotation_select,
                        )
                    )
                compiler_.pre_sql_setup()
                compiler_.column_indices = self.column_indices
                columns = compiler_.columns
                parts.append((compiler_.build_query(columns), compiler_, columns))
            except EmptyResultSet:
                # Omit the empty queryset with UNION and with DIFFERENCE if the
                # first queryset isn't empty.
                if combinator == "union" or (combinator == "difference" and parts):
                    continue
                raise
        # Raise EmptyResultSet if all the combinator queries are empty.
        if not parts:
            raise EmptyResultSet
        # Make the combinator's stages.
        combinator_pipeline = None
        for part, compiler_, columns in parts:
            inner_pipeline = part.get_pipeline()
            # Standardize result fields.
            fields = {}
            # When a .count() is called, the main_query_field has length 1
            # otherwise it has the same length as columns.
            for alias, (ref, expr) in zip(main_query_fields, columns, strict=False):
                if isinstance(expr, Col) and expr.alias != compiler_.collection_name:
                    fields[expr.alias] = 1
                else:
                    fields[alias] = f"${ref}" if alias != ref else 1
            # Avoid duplicating the same $project stage when reusing subquery
            # projections.
            if not inner_pipeline or inner_pipeline[-1] != {"$project": fields}:
                inner_pipeline.append({"$project": fields})
            # Combine query with the current combinator pipeline.
            if combinator_pipeline:
                if combinator == "union":
                    combinator_pipeline.append(
                        {
                            "$unionWith": {
                                "coll": compiler_.base_table.table_name,
                                "pipeline": inner_pipeline,
                            }
                        }
                    )
                elif combinator in {"difference", "intersection"}:
                    # Compute INTERSECTION / DIFFERENCE on projected query
                    # results. The RHS query is evaluated once and reduced to a
                    # distinct set of canonical tuples derived from the
                    # projected output. The LHS query is then evaluated and
                    # filtered by tuple membership against that set. This
                    # ensures correct set semantics even when queries apply
                    # complex transformations or nested lookups.
                    tuple_expr = []
                    for alias, expr in self.columns:
                        if isinstance(expr, Col) and expr.alias != self.collection_name:
                            tuple_expr.append(expr.as_mql(self, self.connection, as_expr=True))
                        else:
                            tuple_expr.append(f"${alias}")
                    distinct_tuple_pipeline = [
                        *inner_pipeline,
                        {"$addFields": {"__tuple": tuple_expr}},
                        {"$group": {"_id": "$__tuple"}},
                    ]
                    # INTERSECTION keeps tuples present in the RHS set.
                    predicate = {"$in": [tuple_expr, "$$rhs_set"]}
                    if combinator == "difference":
                        # DIFFERENCE keeps tuples absent from the RHS set.
                        predicate = {"$not": predicate}
                    combinator_pipeline = [
                        # Anchor pipeline execution to a single document so the
                        # RHS is evaluated once.
                        {"$limit": 1},
                        # Materialize the RHS query as a distinct set of
                        # projected tuples.
                        {
                            "$lookup": {
                                "from": compiler_.base_table.table_name,
                                "pipeline": distinct_tuple_pipeline,
                                "as": "__rhs_set",
                            }
                        },
                        # Keep only the materialized RHS set to minimize
                        # carried state.
                        {"$project": {"__rhs_set": 1}},
                        # Evaluate the LHS and filter by tuple membership.
                        {
                            "$lookup": {
                                "from": self.base_table.table_name,
                                "let": {"rhs_set": "$__rhs_set._id"},
                                "pipeline": [
                                    *combinator_pipeline,
                                    {"$match": {"$expr": predicate}},
                                ],
                                "as": "__result",
                            },
                        },
                        # Restore a standard one-document-per-row result
                        # stream.
                        {"$unwind": "$__result"},
                        {"$replaceRoot": {"newRoot": "$__result"}},
                    ]
            else:
                combinator_pipeline = inner_pipeline
        if not self.query.combinator_all:
            ids = defaultdict(dict)
            for alias, expr in self.columns:
                # Unfold foreign fields.
                if isinstance(expr, Col) and expr.alias != self.collection_name:
                    ids[expr.alias][expr.target.column] = expr.as_mql(
                        self, self.connection, as_expr=True
                    )
                else:
                    ids[alias] = f"${alias}"
            # Convert defaultdict to dict so it doesn't appear as
            # "defaultdict(<CLASS 'dict'>, ..." in query logging.
            ids = dict(ids)
            combinator_pipeline.append({"$group": {"_id": ids}})
            projected_fields = {key: f"$_id.{key}" for key in ids}
            combinator_pipeline.append({"$addFields": projected_fields})
            if "_id" not in projected_fields:
                combinator_pipeline.append({"$unset": "_id"})
        return combinator_pipeline

    def _get_pushable_conditions(self):
        """
        Return a dict mapping each alias to a WhereNode holding its pushable
        condition.
        """
        return self._collect_pushable(self.get_where())

    @classmethod
    def _collect_pushable(cls, expr, negated=False):
        if expr is None or isinstance(expr, (NothingNode, ExtraWhere)):
            return {}
        if isinstance(expr, WhereNode):
            # Apply De Morgan: track negation so connectors are flipped
            # when needed.
            negated ^= expr.negated
            pushable_expressions = [
                cls._collect_pushable(sub_expr, negated=negated)
                for sub_expr in expr.children
                if sub_expr is not None
            ]
            operator = expr.connector
            if operator == XOR:
                return {}
            if negated:
                operator = OR if operator == AND else AND
            alias_children = defaultdict(list)
            for pe in pushable_expressions:
                for alias, expressions in pe.items():
                    alias_children[alias].append(expressions)
            # Build per-alias pushable condition nodes.
            if operator == AND:
                return {
                    alias: WhereNode(children=children, negated=False, connector=operator)
                    for alias, children in alias_children.items()
                }
            # Only aliases shared across all branches are pushable for OR.
            shared_alias = (
                set.intersection(*(set(pe) for pe in pushable_expressions))
                if pushable_expressions
                else set()
            )
            return {
                alias: WhereNode(children=children, negated=False, connector=operator)
                for alias, children in alias_children.items()
                if alias in shared_alias
            }
        # A leaf is pushable only when comparing a field to a constant or
        # simple value.
        if isinstance(expr.lhs, Col) and (
            is_constant_value(expr.rhs) or getattr(expr.rhs, "is_simple_column", False)
        ):
            alias = expr.lhs.alias
            expr = WhereNode(children=[expr], negated=negated)
            return {alias: expr}
        return {}

    def get_lookup_pipeline(self):
        result = []
        # To improve join performance, push conditions (filters) from the
        # WHERE ($match) clause to the JOIN ($lookup) clause.
        pushed_filters = self._get_pushable_conditions()
        for alias in tuple(self.query.alias_map):
            if not self.query.alias_refcount[alias] or self.collection_name == alias:
                continue
            result += self.query.alias_map[alias].as_mql(
                self, self.connection, pushed_filters.get(alias)
            )
        return result

    def _get_aggregate_expressions(self, expr):
        return self._get_all_expressions_of_type(expr, Aggregate)

    def _get_search_expressions(self, expr):
        return self._get_all_expressions_of_type(expr, SearchExpression)

    def _get_all_expressions_of_type(self, expr, target_type):
        stack = [expr]
        while stack:
            expr = stack.pop()
            if isinstance(expr, target_type):
                yield expr
            elif hasattr(expr, "get_source_expressions"):
                stack.extend(expr.get_source_expressions())

    def get_project_fields(self, columns=None, ordering=None, force_expression=False):
        if not columns:
            return {}
        fields = defaultdict(dict)
        for name, expr in columns + (ordering or ()):
            collection = expr.alias if isinstance(expr, Col) else None
            try:
                fields[collection][name] = (
                    1
                    # For brevity/simplicity, project {"field_name": 1}
                    # instead of {"field_name": "$field_name"}.
                    if isinstance(expr, Col) and name == expr.target.column and not force_expression
                    else expr.as_mql(self, self.connection, as_expr=True)
                )
            except EmptyResultSet:
                empty_result_set_value = getattr(expr, "empty_result_set_value", NotImplemented)
                value = (
                    False if empty_result_set_value is NotImplemented else empty_result_set_value
                )
                fields[collection][name] = Value(value).as_mql(self, self.connection, as_expr=True)
            except FullResultSet:
                fields[collection][name] = Value(True).as_mql(self, self.connection, as_expr=True)
        # Annotations (stored in None) and the main collection's fields
        # should appear in the top-level of the fields dict.
        fields.update(fields.pop(None, {}))
        fields.update(fields.pop(self.collection_name, {}))
        # Convert defaultdict to dict so it doesn't appear as
        # "defaultdict(<CLASS 'dict'>, ..." in query logging.
        return dict(fields)

    def _get_ordering(self):
        """
        Process the query's OrderBy objects and return:
        - A tuple of ('field_name': Col/Expression, ...)
        - A bson.SON mapping to pass to $sort.
        - A tuple of ('field_name': Expression, ...) for expressions that need
          to be added to extra_fields.
        """
        fields = []
        sort_ordering = SON()
        extra_fields = []
        idx = itertools.count(start=1)
        for order in self.order_by_objs or []:
            if isinstance(order.expression, Col):
                field_name = order.as_mql(self, self.connection)
                fields.append((order.expression.target.column, order.expression))
            elif isinstance(order.expression, Ref):
                field_name = order.as_mql(self, self.connection)
            else:
                field_name = f"__order{next(idx)}"
                fields.append((field_name, order.expression))
            # If the expression is ordered by NULLS FIRST or NULLS LAST,
            # add a field for sorting that's 1 if null or 0 if not.
            if order.nulls_first or order.nulls_last:
                null_fieldname = f"__order{next(idx)}"
                condition = When(IsNull(order.expression, True), then=Value(1))
                extra_fields.append((null_fieldname, Case(condition, default=Value(0))))
                sort_ordering[null_fieldname] = DESCENDING if order.nulls_first else ASCENDING
            sort_ordering[field_name] = DESCENDING if order.descending else ASCENDING
        return tuple(fields), sort_ordering, tuple(extra_fields)

    def get_where(self):
        return getattr(self, "where", self.query.where)

    def set_where(self, value):
        self.where = value

    def explain_query(self):
        # Validate format (none supported) and options.
        options = self.connection.ops.explain_query_prefix(
            self.query.explain_info.format,
            **self.query.explain_info.options,
        )
        # Build the query pipeline.
        self.pre_sql_setup()
        query = self.build_query(self.get_project_columns(self.columns))
        pipeline = query.get_pipeline()
        # Explain the pipeline.
        kwargs = {}
        for option in self.connection.ops.explain_options:
            if value := options.get(option):
                kwargs[option] = value
        explain = self.connection.database.command(
            "explain",
            {"aggregate": self.collection_name, "pipeline": pipeline, "cursor": {}},
            **kwargs,
        )
        return [json_util.dumps(explain, indent=4, ensure_ascii=False)]

    def as_sql(self, with_limits=True, with_col_aliases=False):
        self.pre_sql_setup()
        pipeline = self.build_query(self.get_project_columns(self.columns)).get_pipeline()
        return f"db.{self.collection_name}.aggregate({pipeline})", []


class SQLInsertCompiler(SQLCompiler):
    def execute_sql(self, returning_fields=None):
        self.pre_sql_setup()
        objs = []
        for obj in self.query.objs:
            field_values = {}
            for field in self.query.fields:
                value = field.get_db_prep_save(
                    getattr(obj, field.attname)
                    if self.query.raw
                    else field.pre_save(obj, obj._state.adding),
                    connection=self.connection,
                )
                if value is None and not field.null and not field.primary_key:
                    raise IntegrityError(
                        f"You can't set {field.name} (a non-nullable field) to None."
                    )

                field_values[field.column] = value
            objs.append(field_values)
        return self.insert(objs, returning_fields=returning_fields)

    @wrap_database_errors
    def insert(self, docs, returning_fields=None):
        """Store a list of documents using field columns as element names."""
        inserted_ids = self.collection.insert_many(
            docs, session=self.connection.session
        ).inserted_ids
        return [(x,) for x in inserted_ids] if returning_fields else []

    @cached_property
    def collection_name(self):
        return self.query.get_meta().db_table


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):
    def execute_sql(self, result_type=MULTI):
        try:
            query = self.build_query()
        except EmptyResultSet:
            return 0
        else:
            return query.delete()

    def check_query(self):
        super().check_query()
        if not self.single_alias:
            raise NotSupportedError(
                "Cannot use QuerySet.delete() when querying across multiple collections on MongoDB."
            )

    def get_where(self):
        return self.query.where

    @cached_property
    def collection_name(self):
        return self.query.base_table


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):
    def execute_sql(self, result_type):
        """
        Execute the specified update. Return the number of rows affected by
        the primary update query. The "primary update query" is the first
        non-empty query that is executed. Row counts for any subsequent,
        related queries are not available.
        """
        self.pre_sql_setup()
        values = {}
        update_has_expression = False
        for field, _, value in self.query.values:
            if hasattr(value, "resolve_expression"):
                update_has_expression = True
                value = value.resolve_expression(self.query, allow_joins=False, for_save=True)
                if value.contains_aggregate:
                    raise FieldError(
                        f"Aggregate functions are not allowed in this query ({field.name}={value})."
                    )
                if value.contains_over_clause:
                    raise FieldError(
                        f"Window expressions are not allowed in this query ({field.name}={value})."
                    )
            elif hasattr(value, "prepare_database_save"):
                if field.remote_field:
                    value = value.prepare_database_save(field)
                elif not getattr(field, "stores_model_instance", False):
                    raise TypeError(
                        f"Tried to update field {field} with a model "
                        f"instance, {value!r}. Use a value compatible with "
                        f"{field.__class__.__name__}."
                    )
            prepared = field.get_db_prep_save(value, connection=self.connection)
            if is_direct_value(value):
                # Encrypted updates don't use aggregation expression syntax and
                # thus must not be escaped.
                if not self.connection.auto_encryption_opts:
                    prepared = {"$literal": prepared}
            else:
                prepared = prepared.as_mql(self, self.connection, as_expr=True)
            values[field.column] = prepared
        try:
            criteria = self.build_query().match_mql
        except EmptyResultSet:
            return 0
        is_empty = not bool(values)
        update_expr = [{"$set": values}]
        if self.connection.auto_encryption_opts:
            if update_has_expression:
                raise NotSupportedError(
                    "Expressions in update queries are not allowed with Queryable Encryption."
                )
            # Pipelines in updates are not allowed with Queryable Encryption.
            update_expr = update_expr[0]
        rows = 0 if is_empty else self.update(criteria, update_expr)
        for query in self.query.get_related_updates():
            aux_rows = query.get_compiler(self.using).execute_sql(result_type)
            if is_empty and aux_rows:
                rows = aux_rows
                is_empty = False
        return rows

    @wrap_database_errors
    def update(self, criteria, pipeline):
        # If "_id" is in the criteria and is an ObjectId, the update will match
        # one document (and is most likely model update). Use update_one()
        # since update_many() isn't supported by Queryable Encryption.
        update_method = (
            "update_one"
            if "_id" in criteria and isinstance(criteria["_id"], ObjectId)
            else "update_many"
        )
        return getattr(self.collection, update_method)(
            criteria, pipeline, session=self.connection.session
        ).matched_count

    def check_query(self):
        super().check_query()
        if len([a for a in self.query.alias_map if self.query.alias_refcount[a]]) > 1:
            raise NotSupportedError(
                "Cannot use QuerySet.update() when querying across multiple collections on MongoDB."
            )

    def get_where(self):
        return self.query.where

    @cached_property
    def collection_name(self):
        return self.query.base_table


class SQLAggregateCompiler(SQLCompiler):
    def build_query(self, columns=None):
        query = self.query_class(self)
        query.project_fields = self.get_project_fields(tuple(self.annotations.items()))
        compiler = self.query.inner_query.get_compiler(
            self.using,
            elide_empty=self.elide_empty,
        )
        compiler.pre_sql_setup(with_col_aliases=False)
        subquery = compiler.build_query(self.get_project_columns(compiler.columns))
        query.subqueries = [subquery]
        return query

    def _make_result(self, result, columns=None):
        return [result[k] for k in self.query.annotation_select]



"""Not a public API."""

from bson import SON, Decimal128, ObjectId


class MongoTestCaseMixin:
    maxDiff = None
    query_types = {"SON": SON, "ObjectId": ObjectId, "Decimal128": Decimal128}

    def assertAggregateQuery(self, query, expected_collection, expected_pipeline):
        """
        Assert that the logged query is equal to:
            db.{expected_collection}.aggregate({expected_pipeline})
        """
        prefix, pipeline = query.split("(", 1)
        _, collection, operator = prefix.split(".")
        self.assertEqual(operator, "aggregate")
        self.assertEqual(collection, expected_collection)
        self.assertEqual(eval(pipeline[:-1], self.query_types, {}), expected_pipeline)  # noqa: S307

    def assertInsertQuery(self, query, expected_collection, expected_documents):
        """
        Assert that the logged query is equal to:
            db.{expected_collection}.insert_many({expected_documents})
        """
        prefix, pipeline = query.split("(", 1)
        _, collection, operator = prefix.split(".")
        self.assertEqual(operator, "insert_many")
        self.assertEqual(collection, expected_collection)
        self.assertEqual(eval(pipeline[:-1], self.query_types), expected_documents)  # noqa: S307

    def assertUpdateQuery(self, query, expected_collection, expected_condition, expected_set):
        """
        Assert that the logged query is equal to:
            db.{expected_collection}.update_many({expected_condition}, {expected_set})
        """
        prefix, pipeline = query.split("(", 1)
        _, collection, operator = prefix.split(".")
        self.assertEqual(operator, "update_many")
        self.assertEqual(collection, expected_collection)
        condition, set_expression = eval(pipeline[:-1], self.query_types, {})  # noqa: S307
        self.assertEqual(condition, expected_condition)
        self.assertEqual(set_expression, expected_set)

    def assertUpdateOneQuery(self, query, expected_collection, expected_condition, expected_set):
        """
        Assert that the logged query is equal to:
            db.{expected_collection}.update_one({expected_condition}, {expected_set})
        """
        prefix, pipeline = query.split("(", 1)
        _, collection, operator = prefix.split(".")
        self.assertEqual(operator, "update_one")
        self.assertEqual(collection, expected_collection)
        condition, set_expression = eval(pipeline[:-1], self.query_types, {})  # noqa: S307
        self.assertEqual(condition, expected_condition)
        self.assertEqual(set_expression, expected_set)



# Settings for django_mongodb_backend/tests when encryption is supported.
import os

from mongodb_settings import *  # noqa: F403
from pymongo.encryption import AutoEncryptionOpts

DATABASES["encrypted"] = {  # noqa: F405
    "ENGINE": "django_mongodb_backend",
    "NAME": "djangotests_encrypted",
    "OPTIONS": {
        "auto_encryption_opts": AutoEncryptionOpts(
            key_vault_namespace="djangotests_encrypted.__keyVault",
            kms_providers={"local": {"key": os.urandom(96)}},
            crypt_shared_lib_path=os.environ["CRYPT_SHARED_LIB_PATH"],
            crypt_shared_lib_required=True,
        ),
        "directConnection": True,
    },
}


class EncryptedRouter:
    def db_for_read(self, model, **hints):
        # All models in the encryption_ app use the encrypted database.
        if model._meta.app_label == "encryption_":
            return "encrypted"
        return None

    db_for_write = db_for_read

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        # Create the encryption_ app's models only in the encrypted database.
        if app_label == "encryption_":
            return db == "encrypted"
        # Don't create other apps' models in the encrypted database.
        if db == "encrypted":
            return False
        return None


DATABASE_ROUTERS.append(EncryptedRouter())  # noqa: F405



import unittest
from datetime import date
from operator import attrgetter

from django.core.exceptions import FieldDoesNotExist
from django.db import connection, models
from django.db.models.expressions import Value
from django.db.models.functions import Concat
from django.test import SimpleTestCase, TestCase
from django.test.utils import CaptureQueriesContext, isolate_apps

from django_mongodb_backend.fields import ArrayField, EmbeddedModelArrayField
from django_mongodb_backend.models import EmbeddedModel
from django_mongodb_backend.test import MongoTestCaseMixin

from .models import Artifact, Audit, Exhibit, Movie, Restoration, Review, Section, Tour


class MethodTests(SimpleTestCase):
    def test_deconstruct(self):
        field = EmbeddedModelArrayField("Data", null=True)
        field.name = "field_name"
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(name, "field_name")
        self.assertEqual(path, "django_mongodb_backend.fields.EmbeddedModelArrayField")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {"embedded_model": "data", "null": True})

    def test_deconstruct_model_class(self):
        field = EmbeddedModelArrayField(Movie)
        *_, kwargs = field.deconstruct()
        self.assertEqual(kwargs["embedded_model"], "model_fields_.movie")

    def test_deconstruct_model_string(self):
        field = EmbeddedModelArrayField("app_label.Model")
        *_, kwargs = field.deconstruct()
        self.assertEqual(kwargs["embedded_model"], "app_label.model")

    def test_size_not_supported(self):
        msg = "EmbeddedModelArrayField does not support size."
        with self.assertRaisesMessage(ValueError, msg):
            EmbeddedModelArrayField("Data", size=1)

    def test_get_db_prep_save_invalid(self):
        msg = "Expected list of <class 'model_fields_.models.Review'> instances, not <class 'int'>."
        with self.assertRaisesMessage(TypeError, msg):
            Movie(reviews=42).save()

    def test_get_db_prep_save_invalid_list(self):
        msg = "Expected instance of type <class 'model_fields_.models.Review'>, not <class 'int'>."
        with self.assertRaisesMessage(TypeError, msg):
            Movie(reviews=[42]).save()


class ModelTests(TestCase):
    def test_save_load(self):
        reviews = [
            Review(title="The best", rating=10),
            Review(title="Mediocre", rating=5),
            Review(title="Horrible", rating=1),
        ]
        Movie.objects.create(title="Lion King", reviews=reviews)
        movie = Movie.objects.get(title="Lion King")
        self.assertEqual(movie.reviews[0].title, "The best")
        self.assertEqual(movie.reviews[0].rating, 10)
        self.assertEqual(movie.reviews[1].title, "Mediocre")
        self.assertEqual(movie.reviews[1].rating, 5)
        self.assertEqual(movie.reviews[2].title, "Horrible")
        self.assertEqual(movie.reviews[2].rating, 1)
        self.assertEqual(len(movie.reviews), 3)

    def test_save_load_null(self):
        movie = Movie.objects.create(title="Lion King")
        movie = Movie.objects.get(title="Lion King")
        self.assertIsNone(movie.reviews)

    def test_missing_field_in_data(self):
        """
        Loading a model with an EmbeddedModelArrayField that has a missing
        subfield (e.g. data not written by Django) that uses a database
        converter (in this case, rating is an IntegerField) doesn't crash.
        """
        Movie.objects.create(title="Lion King", reviews=[Review(title="The best", rating=10)])
        connection.database.model_fields__movie.update_many(
            {}, {"$unset": {"reviews.$[].rating": ""}}
        )
        self.assertIsNone(Movie.objects.first().reviews[0].rating)

    def test_embedded_model_field_respects_db_column(self):
        """
        EmbeddedModel data respects Field.db_column. In this case,
        Review.title has db_column="title_".
        """
        obj = Movie.objects.create(title="Lion King", reviews=[Review(title="Awesome", rating=10)])
        query = connection.database.model_fields__movie.find({"_id": obj.pk})
        self.assertEqual(query[0]["reviews"][0]["title_"], "Awesome")


class QueryingTests(MongoTestCaseMixin, TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.egypt = Exhibit.objects.create(
            name="Ancient Egypt",
            sections=[
                Section(
                    number=1,
                    artifacts=[
                        Artifact(
                            name="Ptolemaic Crown",
                            metadata={
                                "origin": "Egypt",
                            },
                        )
                    ],
                )
            ],
        )
        cls.wonders = Exhibit.objects.create(
            name="Wonders of the Ancient World",
            sections=[
                Section(
                    number=1,
                    artifacts=[
                        Artifact(
                            name="Statue of Zeus",
                            metadata={"location": "Olympia", "height_m": 12},
                        ),
                        Artifact(
                            name="Hanging Gardens",
                        ),
                    ],
                ),
                Section(
                    number=2,
                    artifacts=[
                        Artifact(
                            name="Lighthouse of Alexandria",
                            metadata={"height_m": 100, "built": "3rd century BC"},
                        )
                    ],
                ),
            ],
        )
        cls.new_discoveries = Exhibit.objects.create(
            name="New Discoveries",
            sections=[
                Section(
                    number=2,
                    artifacts=[
                        Artifact(
                            name="Lighthouse of Alexandria",
                            metadata={"height_m": 100, "built": "3rd century BC"},
                        )
                    ],
                )
            ],
            main_section=Section(number=2),
        )
        cls.lost_empires = Exhibit.objects.create(
            name="Lost Empires",
            main_section=Section(
                number=3,
                artifacts=[
                    Artifact(
                        name="Bronze Statue",
                        metadata={"origin": "Pergamon"},
                        restorations=[
                            Restoration(
                                date=date(1998, 4, 15),
                                restored_by="Zacarias",
                            ),
                            Restoration(
                                date=date(2010, 7, 22),
                                restored_by="Vicente",
                            ),
                        ],
                        last_restoration=Restoration(
                            date=date(2010, 7, 22),
                            restored_by="Monzon",
                        ),
                    )
                ],
            ),
        )
        cls.egypt_tour = Tour.objects.create(guide="Amira", exhibit=cls.egypt)
        cls.wonders_tour = Tour.objects.create(guide="Carlos", exhibit=cls.wonders)
        cls.lost_tour = Tour.objects.create(guide="Yelena", exhibit=cls.lost_empires)
        cls.audit_1 = Audit.objects.create(section_number=1, reviewed=True)
        cls.audit_2 = Audit.objects.create(section_number=2, reviewed=True)
        cls.audit_3 = Audit.objects.create(section_number=5, reviewed=False)

    def test_exact(self):
        with self.assertNumQueries(1) as ctx:
            self.assertCountEqual(
                Exhibit.objects.filter(sections__number=1), [self.egypt, self.wonders]
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "model_fields__exhibit",
            [{"$match": {"sections.number": 1}}],
        )

    def test_exact_expr(self):
        with self.assertNumQueries(1) as ctx:
            self.assertCountEqual(
                Exhibit.objects.filter(sections__number=Value(2) - 1), [self.egypt, self.wonders]
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "model_fields__exhibit",
            [
                {
                    "$match": {
                        "$expr": {
                            "$anyElementTrue": {
                                "$ifNull": [
                                    {
                                        "$map": {
                                            "input": "$sections",
                                            "as": "item",
                                            "in": {
                                                "$eq": [
                                                    "$$item.number",
                                                    {
                                                        "$subtract": [
                                                            {"$literal": 2},
                                                            {"$literal": 1},
                                                        ]
                                                    },
                                                ]
                                            },
                                        }
                                    },
                                    [],
                                ]
                            }
                        }
                    }
                }
            ],
        )

    def test_array_index(self):
        with self.assertNumQueries(1) as ctx:
            self.assertCountEqual(
                Exhibit.objects.filter(sections__0__number=1),
                [self.egypt, self.wonders],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "model_fields__exhibit",
            [{"$match": {"sections.0.number": 1}}],
        )

    def test_array_index_expr(self):
        with self.assertNumQueries(1) as ctx:
            self.assertCountEqual(
                Exhibit.objects.filter(sections__0__number=Value(2) - 1),
                [self.egypt, self.wonders],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "model_fields__exhibit",
            [
                {
                    "$match": {
                        "$expr": {
                            "$eq": [
                                {
                                    "$getField": {
                                        "input": {"$arrayElemAt": ["$sections", 0]},
                                        "field": "number",
                                    }
                                },
                                {"$subtract": [{"$literal": 2}, {"$literal": 1}]},
                            ]
                        }
                    }
                }
            ],
        )

    def test_nested_array_index(self):
        with self.assertNumQueries(1) as ctx:
            self.assertCountEqual(
                Exhibit.objects.filter(
                    main_section__artifacts__restorations__0__restored_by="Zacarias"
                ),
                [self.lost_empires],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "model_fields__exhibit",
            [{"$match": {"main_section.artifacts.restorations.0.restored_by": "Zacarias"}}],
        )

    def test_nested_array_index_expr(self):
        with self.assertNumQueries(1) as ctx:
            self.assertCountEqual(
                Exhibit.objects.filter(
                    main_section__artifacts__restorations__0__restored_by=Concat(
                        Value("Z"), Value("acarias")
                    )
                ),
                [self.lost_empires],
            )
        self.assertAggregateQuery(
            ctx.captured_queries[0]["sql"],
            "model_fields__exhibit",
            [
                {
                    "$match": {
                        "$expr": {
                            "$anyElementTrue": {
                                "$ifNull": [
                                    {
                                        "$map": {
                                            "input": {
                                                "$getField": {
                                                    "input": "$main_section",
                                                    "field": "artifacts",
                                                }
                                            },
                                            "as": "item",
                                            "in": {
                                                "$eq": [
                                                    {
                                                        "$getField": {
                                                            "input": {
                                                                "$arrayElemAt": [
                                                                    "$$item.restorations",
                                                                    0,
                                                                ]
                                                            },
                                                            "field": "restored_by",
                                                        }
                                                    },
                                                    {
                                                        "$concat": [
                                                            {
                                                                "$ifNull": [
                                                                    {"$literal": "Z"},
                                                                    {"$literal": ""},
                                                                ]
                                                            },
                                                            {
                                                                "$ifNull": [
                                                                    {"$literal": "acarias"},
                                                                    {"$literal": ""},
                                                                ]
                                                            },
                                                        ]
                                                    },
                                                ]
                                            },
                                        }
                                    },
                                    [],
                                ]
                            }
                        }
                    }
                }
            ],
        )

    def test_array_slice(self):
        self.assertSequenceEqual(
            Exhibit.objects.filter(sections__0_1__number=2), [self.new_discoveries]
        )

    def test_filter_unsupported_lookups_in_json(self):
        """Unsupported lookups can be used as keys in a JSONField."""
        for lookup in ["contains", "range"]:
            kwargs = {f"main_section__artifacts__metadata__origin__{lookup}": ["Pergamon", "Egypt"]}
            with CaptureQueriesContext(connection) as captured_queries:
                self.assertCountEqual(Exhibit.objects.filter(**kwargs), [])
            query = captured_queries[0]["sql"]
            self.assertAggregateQuery(
                query,
                "model_fields__exhibit",
                [
                    {
                        "$match": {
                            f"main_section.artifacts.metadata.origin.{lookup}": [
                                "Pergamon",
                                "Egypt",
                            ]
                        }
                    }
                ],
            )

    def test_len(self):
        self.assertCountEqual(Exhibit.objects.filter(sections__len=10), [])
        self.assertCountEqual(
            Exhibit.objects.filter(sections__len=1),
            [self.egypt, self.new_discoveries],
        )
        # Nested EMF
        self.assertCountEqual(
            Exhibit.objects.filter(main_section__artifacts__len=1), [self.lost_empires]
        )
        self.assertCountEqual(Exhibit.objects.filter(main_section__artifacts__len=2), [])
        # Nested Indexed Array
        self.assertCountEqual(Exhibit.objects.filter(sections__0__artifacts__len=2), [self.wonders])
        self.assertCountEqual(Exhibit.objects.filter(sections__0__artifacts__len=0), [])
        self.assertCountEqual(Exhibit.objects.filter(sections__1__artifacts__len=1), [self.wonders])

    def test_in(self):
        self.assertCountEqual(Exhibit.objects.filter(sections__number__in=[10]), [])
        self.assertCountEqual(
            Exhibit.objects.filter(sections__number__in=[1]),
            [self.egypt, self.wonders],
        )
        self.assertCountEqual(
            Exhibit.objects.filter(sections__number__in=[2]),
            [self.new_discoveries, self.wonders],
        )
        self.assertCountEqual(Exhibit.objects.filter(sections__number__in=[3]), [])

    def test_iexact(self):
        self.assertCountEqual(
            Exhibit.objects.filter(sections__artifacts__0__name__iexact="lightHOuse of aLexandriA"),
            [self.new_discoveries, self.wonders],
        )

    def test_gt(self):
        self.assertCountEqual(
            Exhibit.objects.filter(sections__number__gt=1),
            [self.new_discoveries, self.wonders],
        )

    def test_gte(self):
        self.assertCountEqual(
            Exhibit.objects.filter(sections__number__gte=1),
            [self.egypt, self.new_discoveries, self.wonders],
        )

    def test_lt(self):
        self.assertCountEqual(
            Exhibit.objects.filter(sections__number__lt=2), [self.egypt, self.wonders]
        )

    def test_lte(self):
        self.assertCountEqual(
            Exhibit.objects.filter(sections__number__lte=2),
            [self.egypt, self.wonders, self.new_discoveries],
        )

    def test_querying_array_not_allowed(self):
        msg = (
            "Lookups aren't supported on EmbeddedModelArrayField. "
            "Try querying one of its embedded fields instead."
        )
        with self.assertRaisesMessage(ValueError, msg):
            Exhibit.objects.filter(sections=10).first()

        with self.assertRaisesMessage(ValueError, msg):
            Exhibit.objects.filter(sections__0_1=10).first()

    def test_invalid_field(self):
        msg = "Section has no field named 'section'"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Exhibit.objects.filter(sections__section__in=[10]).first()

    def test_invalid_nested_field(self):
        msg = "Artifact has no field named 'xx'"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Exhibit.objects.filter(sections__artifacts__xx=10).first()

    def test_invalid_lookup(self):
        msg = "Unsupported lookup 'return' for EmbeddedModelArrayField of 'IntegerField'"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Exhibit.objects.filter(sections__number__return=3)

    def test_unsupported_lookup(self):
        msg = "Unsupported lookup 'range' for EmbeddedModelArrayField of 'IntegerField'"
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Exhibit.objects.filter(sections__number__range=[10])

    def test_missing_lookup_suggestions(self):
        msg = (
            "Unsupported lookup 'ltee' for EmbeddedModelArrayField of 'IntegerField', "
            "perhaps you meant lte or lt?"
        )
        with self.assertRaisesMessage(FieldDoesNotExist, msg):
            Exhibit.objects.filter(sections__number__ltee=3)

    def test_nested_lookup(self):
        msg = "Cannot perform multiple levels of array traversal in a query."
        with self.assertRaisesMessage(ValueError, msg):
            Exhibit.objects.filter(sections__artifacts__name="")

    def test_foreign_field_exact(self):
        """Querying from a foreign key to an EmbeddedModelArrayField."""
        with self.assertNumQueries(1) as ctx:
            qs = Tour.objects.filter(exhibit__sections__number=1)
            self.assertCountEqual(qs, [self.egypt_tour, self.wonders_tour])
        self.assertNotIn("anyElementTrue", ctx.captured_queries[0]["sql"])

    def test_foreign_field_exact_expr(self):
        with self.assertNumQueries(1) as ctx:
            qs = Tour.objects.filter(exhibit__sections__number=Value(2) - Value(1))
            self.assertCountEqual(qs, [self.egypt_tour, self.wonders_tour])
        self.assertIn("anyElementTrue", ctx.captured_queries[0]["sql"])

    def test_foreign_field_with_slice(self):
        qs = Tour.objects.filter(exhibit__sections__0_2__number__in=[1, 2])
        self.assertCountEqual(qs, [self.wonders_tour, self.egypt_tour])

    def test_subquery_numeric_lookups(self):
        subquery = Audit.objects.filter(
            section_number__in=models.OuterRef("sections__number")
        ).values("section_number")[:1]
        tests = [
            ("exact", [self.egypt, self.new_discoveries, self.wonders]),
            ("lt", []),
            ("lte", [self.egypt, self.new_discoveries, self.wonders]),
            ("gt", [self.wonders]),
            ("gte", [self.egypt, self.new_discoveries, self.wonders]),
        ]
        for lookup, expected in tests:
            with self.subTest(lookup=lookup):
                kwargs = {f"sections__number__{lookup}": subquery}
                self.assertCountEqual(Exhibit.objects.filter(**kwargs), expected)

    def test_subquery_in_lookup(self):
        subquery = Audit.objects.filter(reviewed=True).values_list("section_number", flat=True)
        result = Exhibit.objects.filter(sections__number__in=subquery)
        self.assertCountEqual(result, [self.wonders, self.new_discoveries, self.egypt])

    def test_subquery_empty_in_lookup(self):
        subquery = Audit.objects.filter(section_number=10).values_list("section_number", flat=True)
        result = Exhibit.objects.filter(sections__number__in=subquery)
        self.assertCountEqual(result, [])

    def test_array_as_rhs(self):
        result = Exhibit.objects.filter(main_section__number__in=models.F("sections__number"))
        self.assertCountEqual(result, [self.new_discoveries])

    def test_array_annotation_lookup(self):
        result = Exhibit.objects.annotate(section_numbers=models.F("main_section__number")).filter(
            section_numbers__in=models.F("sections__number")
        )
        self.assertCountEqual(result, [self.new_discoveries])

    def test_array_as_rhs_for_arrayfield_lookups(self):
        tests = [
            ("exact", [self.wonders]),
            ("lt", [self.new_discoveries]),
            ("lte", [self.wonders, self.new_discoveries]),
            ("gt", [self.egypt, self.lost_empires]),
            ("gte", [self.egypt, self.wonders, self.lost_empires]),
            ("overlap", [self.egypt, self.wonders, self.new_discoveries]),
            ("contained_by", [self.wonders]),
            ("contains", [self.egypt, self.wonders, self.new_discoveries, self.lost_empires]),
        ]
        for lookup, expected in tests:
            with self.subTest(lookup=lookup):
                kwargs = {f"section_numbers__{lookup}": models.F("sections__number")}
                result = Exhibit.objects.annotate(
                    section_numbers=Value(
                        [1, 2], output_field=ArrayField(base_field=models.IntegerField())
                    )
                ).filter(**kwargs)
                self.assertCountEqual(result, expected)

    @unittest.expectedFailure
    def test_array_annotation_index(self):
        # Slicing and indexing over an annotated EmbeddedModelArrayField would
        # require a refactor of annotation handling.
        result = Exhibit.objects.annotate(section_numbers=models.F("sections__number")).filter(
            section_numbers__0=1
        )
        self.assertCountEqual(result, [self.new_discoveries, self.egypt])

    def test_array_annotation(self):
        qs = Exhibit.objects.annotate(section_numbers=models.F("sections__number")).order_by("name")
        self.assertQuerySetEqual(qs, [[1], [], [2], [1, 2]], attrgetter("section_numbers"))


@isolate_apps("model_fields_")
class CheckTests(SimpleTestCase):
    def test_no_relational_fields(self):
        class Target(EmbeddedModel):
            key = models.ForeignKey("MyModel", models.CASCADE)

        class MyModel(models.Model):
            field = EmbeddedModelArrayField(Target)

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.fields.array.E001")
        msg = errors[0].msg
        self.assertEqual(
            msg,
            "Base field for array has errors:\n    "
            "Embedded models cannot have relational fields (Target.key is a ForeignKey). "
            "(mongodb.fields.embedded_model.E001)",
        )

    def test_embedded_model_subclass(self):
        class Target(models.Model):
            pass

        class MyModel(models.Model):
            field = EmbeddedModelArrayField(Target)

        errors = MyModel().check()
        self.assertEqual(len(errors), 1)
        self.assertEqual(errors[0].id, "mongodb.fields.array.E001")
        msg = errors[0].msg
        self.assertEqual(
            msg,
            "Base field for array has errors:\n    "
            "Embedded models must be a subclass of "
            "django_mongodb_backend.models.EmbeddedModel. "
            "(mongodb.fields.embedded_model.E002)",
        )



# Settings for django_mongodb_backend/tests with AWS Key Management System.
import os

from encrypted_settings import *  # noqa: F403
from pymongo.encryption import AutoEncryptionOpts

DATABASES["encrypted"] = {  # noqa: F405
    "ENGINE": "django_mongodb_backend",
    "NAME": "djangotests_encrypted",
    "OPTIONS": {
        "auto_encryption_opts": AutoEncryptionOpts(
            key_vault_namespace="djangotests_encrypted.__keyVault",
            kms_providers={
                "aws": {
                    "accessKeyId": os.environ["AWS_ACCESS_KEY_ID"],
                    "secretAccessKey": os.environ["AWS_SECRET_ACCESS_KEY"],
                }
            },
            crypt_shared_lib_path=os.environ["CRYPT_SHARED_LIB_PATH"],
            crypt_shared_lib_required=True,
        ),
    },
    "KMS_CREDENTIALS": {
        "aws": {
            "key": os.environ["AWS_KMS_ARN"],
            "region": os.environ["AWS_DEFAULT_REGION"],
        }
    },
}



from .expressions import Remove
from .search import (
    CombinedSearchExpression,
    CompoundExpression,
    SearchAutocomplete,
    SearchEquals,
    SearchExists,
    SearchGeoShape,
    SearchGeoWithin,
    SearchIn,
    SearchMoreLikeThis,
    SearchPhrase,
    SearchQueryString,
    SearchRange,
    SearchRegex,
    SearchScoreOption,
    SearchText,
    SearchVector,
    SearchWildcard,
)

__all__ = [
    "CombinedSearchExpression",
    "CompoundExpression",
    "Remove",
    "SearchAutocomplete",
    "SearchEquals",
    "SearchExists",
    "SearchGeoShape",
    "SearchGeoWithin",
    "SearchIn",
    "SearchMoreLikeThis",
    "SearchPhrase",
    "SearchQueryString",
    "SearchRange",
    "SearchRegex",
    "SearchScoreOption",
    "SearchText",
    "SearchVector",
    "SearchWildcard",
]



# Settings for django_mongodb_backend/tests when encryption isn't supported.
from django_settings import *  # noqa: F403

DATABASES["encrypted"] = {}  # noqa: F405
DATABASE_ROUTERS = ["django_mongodb_backend.routers.MongoRouter"]



from django.core import checks
from django.db import NotSupportedError, models
from django.test import SimpleTestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.models import EmbeddedModel

from .models import Embed


class TestMethods(SimpleTestCase):
    def test_save(self):
        e = Embed()
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be saved."):
            e.save()

    def test_delete(self):
        e = Embed()
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be deleted."):
            e.delete()


@isolate_apps("models_")
class TestChecks(SimpleTestCase):
    def test_db_index(self):
        class Target(EmbeddedModel):
            foo = models.IntegerField(db_index=True)

        errors = Target().check()
        self.assertEqual(
            errors,
            [
                checks.Warning(
                    "Using db_index=True on embedded fields is deprecated "
                    "in favor of using EmbeddedFieldIndex in Meta.indexes "
                    "on the top-level model.",
                    id="mongodb.fields.embedded_model.W004",
                    obj=Target._meta.get_field("foo"),
                )
            ],
        )

    def test_unique(self):
        class Target(EmbeddedModel):
            foo = models.IntegerField(unique=True)

        errors = Target().check()
        self.assertEqual(
            errors,
            [
                checks.Warning(
                    "Using unique=True on embedded fields is deprecated "
                    "in favor of using EmbeddedFieldUniqueConstraint in "
                    "Meta.constraints on the top-level model.",
                    id="mongodb.fields.embedded_model.W005",
                    obj=Target._meta.get_field("foo"),
                )
            ],
        )

    def test_constraints(self):
        class Target(EmbeddedModel):
            foo = models.IntegerField()

            class Meta:
                constraints = [models.UniqueConstraint(fields=["foo"], name="name")]

        errors = Target().check()
        self.assertEqual(
            errors,
            [
                checks.Warning(
                    "Using Meta.constraints on embedded models is deprecated "
                    "in favor of using EmbeddedFieldUniqueConstraint in "
                    "Meta.constraints on the top-level model.",
                    id="mongodb.fields.embedded_model.W006",
                    obj=Target,
                )
            ],
        )

    def test_indexes(self):
        class Target(EmbeddedModel):
            foo = models.IntegerField()

            class Meta:
                indexes = [models.Index(fields=["foo"])]

        errors = Target().check()
        self.assertEqual(
            errors,
            [
                checks.Warning(
                    "Using Meta.indexes on embedded models is deprecated in "
                    "favor of using EmbeddedFieldIndex in Meta.indexes on the "
                    "top-level model.",
                    id="mongodb.fields.embedded_model.W007",
                    obj=Target,
                )
            ],
        )

    def test_unique_together(self):
        class Target(EmbeddedModel):
            a = models.IntegerField()
            b = models.IntegerField()

            class Meta:
                unique_together = ["a", "b"]

        errors = Target().check()
        self.assertEqual(
            errors,
            [
                checks.Warning(
                    "Using Meta.unique_together on embedded models is "
                    "deprecated in favor of using EmbeddedFieldUniqueConstraint "
                    "in Meta.constraints on the top-level model.",
                    id="mongodb.fields.embedded_model.W008",
                    obj=Target,
                )
            ],
        )


class TestManagerMethods(SimpleTestCase):
    def test_all(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be queried."):
            Embed.objects.all()

    def test_get(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be queried."):
            Embed.objects.get(foo="bar")

    def test_filter(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be queried."):
            Embed.objects.filter(foo="bar")

    def test_create(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be created."):
            Embed.objects.create(foo="bar")

    def test_update(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be updated."):
            Embed.objects.update(foo="bar")

    def test_delete(self):
        with self.assertRaisesMessage(NotSupportedError, "EmbeddedModels cannot be deleted."):
            Embed.objects.delete()

    def test_get_or_create(self):
        msg = "'EmbeddedModelManager' object has no attribute 'get_or_create'"
        with self.assertRaisesMessage(AttributeError, msg):
            Embed.objects.get_or_create()



from functools import reduce, wraps
from operator import add as add_operator

from django.core.exceptions import EmptyResultSet, FullResultSet
from django.db import DatabaseError, IntegrityError, NotSupportedError
from django.db.models.expressions import Case, Col, When
from django.db.models.fields.related import ForeignKey
from django.db.models.functions import Mod
from django.db.models.lookups import Exact
from django.db.models.sql.constants import INNER
from django.db.models.sql.datastructures import Join
from django.db.models.sql.where import AND, OR, XOR, ExtraWhere, NothingNode, WhereNode
from pymongo.errors import BulkWriteError, DuplicateKeyError, PyMongoError


def wrap_database_errors(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BulkWriteError as e:
            if "E11000 duplicate key error" in str(e):
                raise IntegrityError(str(e)) from e
            raise
        except DuplicateKeyError as e:
            raise IntegrityError(str(e)) from e
        except PyMongoError as e:
            raise DatabaseError(str(e)) from e

    return wrapper


class MongoQuery:
    """
    Compilers build a MongoQuery when they want to fetch some data. They work
    by first allowing sql.compiler.SQLCompiler to partly build a sql.Query,
    constructing a MongoQuery query on top of it, and then iterating over its
    results.

    This class provides a framework for converting the SQL constraint tree
    built by Django to a "representation" more suitable for MongoDB.
    """

    def __init__(self, compiler):
        self.compiler = compiler
        self.query = compiler.query
        self.ordering = []
        self.match_mql = {}
        self.subqueries = None
        self.lookup_pipeline = None
        self.project_fields = None
        self.aggregation_pipeline = compiler.aggregation_pipeline
        self.search_pipeline = compiler.search_pipeline
        self.extra_fields = None
        self.combinator_pipeline = None
        # $lookup stage that encapsulates the pipeline for performing a nested
        # subquery.
        self.subquery_lookup = None
        self.needs_wrap_aggregation = compiler.needs_wrap_aggregation

    def __repr__(self):
        return f"<MongoQuery: {self.match_mql!r} ORDER {self.ordering!r}>"

    @wrap_database_errors
    def delete(self):
        """Execute a delete query."""
        if self.compiler.subqueries:
            raise NotSupportedError("Cannot use QuerySet.delete() when a subquery is required.")
        return self.compiler.collection.delete_many(
            self.match_mql, session=self.compiler.connection.session
        ).deleted_count

    @wrap_database_errors
    def get_cursor(self):
        """
        Return a pymongo CommandCursor that can be iterated on to give the
        results of the query.
        """
        return self.compiler.collection.aggregate(
            self.get_pipeline(), session=self.compiler.connection.session
        )

    def get_pipeline(self):
        pipeline = []
        if self.search_pipeline:
            pipeline.extend(self.search_pipeline)
        if self.lookup_pipeline:
            pipeline.extend(self.lookup_pipeline)
        for query in self.subqueries or ():
            pipeline.extend(query.get_pipeline())
        if self.match_mql:
            pipeline.append({"$match": self.match_mql})
        if self.aggregation_pipeline:
            pipeline.extend(self.aggregation_pipeline)
        if self.needs_wrap_aggregation:
            if self.compiler.connection.client_encryption:
                # Automatic encryption doesn't support $unionWith, so use
                # $collStats instead (which may perform a bit worse).
                pipeline = [
                    {"$collStats": {}},
                    {
                        "$lookup": {
                            "from": self.compiler.collection_name,
                            "as": "wrapped",
                            "pipeline": pipeline,
                        }
                    },
                    {
                        "$replaceWith": {
                            "$cond": [{"$eq": ["$wrapped", []]}, {}, {"$first": "$wrapped"}]
                        }
                    },
                ]
            else:
                # Add the aggregation stage for queries without a GROUP BY.
                # e.g. SQL equivalent of SELECT avg(col) FROM table
                pipeline.extend(
                    [
                        # Workaround for https://jira.mongodb.org/browse/SERVER-114196:
                        # $$NOW becomes unavailable after $unionWith, so it
                        # must be stored beforehand to ensure it remains
                        # accessible later in the pipeline.
                        {"$addFields": {"__now": "$$NOW"}},
                        # Add an empty extra document to handle default values
                        # on empty results.
                        {"$unionWith": {"pipeline": [{"$documents": [{}]}]}},
                        # Limiting to one document ensures the original result
                        # takes precedence when present, otherwise the injected
                        # empty document is used.
                        {"$limit": 1},
                    ]
                )
        if self.project_fields:
            pipeline.append({"$project": self.project_fields})
        if self.combinator_pipeline:
            pipeline.extend(self.combinator_pipeline)
        if self.extra_fields:
            pipeline.append({"$addFields": self.extra_fields})
        if self.ordering:
            pipeline.append({"$sort": self.ordering})
        if self.query.low_mark > 0:
            pipeline.append({"$skip": self.query.low_mark})
        if self.query.high_mark is not None:
            pipeline.append({"$limit": self.query.high_mark - self.query.low_mark})
        if self.subquery_lookup:
            table_output = self.subquery_lookup["as"]
            pipeline = [
                {"$lookup": {**self.subquery_lookup, "pipeline": pipeline}},
                {
                    "$set": {
                        table_output: {
                            "$cond": {
                                "if": {
                                    "$or": [
                                        {"$eq": [{"$type": f"${table_output}"}, "missing"]},
                                        {"$eq": [{"$size": f"${table_output}"}, 0]},
                                    ]
                                },
                                "then": {},
                                "else": {"$arrayElemAt": [f"${table_output}", 0]},
                            }
                        }
                    }
                },
            ]
        return pipeline


def extra_where(self, compiler, connection, as_expr=False):  # noqa: ARG001
    raise NotSupportedError("QuerySet.extra() is not supported on MongoDB.")


def join(self, compiler, connection, pushed_filter_expression=None):
    """
    Generate a MongoDB $lookup stage for a join.

    `pushed_filter_expression` is a Where expression involving fields from the
    joined collection which can be pushed from the WHERE ($match) clause to the
    JOIN ($lookup) clause to improve performance.
    """
    parent_template = "parent__field__"

    def _get_reroot_replacements(expression):
        if not expression:
            return None
        columns = []
        for expr in expression.leaves():
            # Determine whether the column needs to be transformed or rerouted
            # as part of the subquery.
            for hand_side in ["lhs", "rhs"]:
                hand_side_value = getattr(expr, hand_side, None)
                if isinstance(hand_side_value, Col):
                    # If the column is not part of the joined table, add it to
                    # lhs_fields.
                    if hand_side_value.alias != self.table_alias:
                        pos = len(lhs_fields)
                        lhs_fields.append(
                            hand_side_value.as_mql(compiler, connection, as_expr=True)
                        )
                    else:
                        pos = None
                    columns.append((hand_side_value, pos))
        # Replace columns in the extra conditions with new column references
        # based on their rerouted positions in the join pipeline.
        replacements = {}
        for col, parent_pos in columns:
            target = col.target.clone()
            target.remote_field = col.target.remote_field
            column_target = Col(compiler.collection_name, target)
            if parent_pos is not None:
                column_target.is_simple_column = False
                target_col = f"${parent_template}{parent_pos}"
                column_target.target.db_column = target_col
                column_target.target.set_attributes_from_name(target_col)
            else:
                column_target.target = col.target
            replacements[col] = column_target
        return replacements

    lookup_pipeline = []
    lhs_fields = []
    rhs_fields = []
    local_field = None
    foreign_field = None
    # Add a join condition for each pair of joining fields.
    for lhs, rhs in self.join_fields:
        lhs_prepared, rhs_prepared = connection.ops.prepare_join_on_clause(
            self.parent_alias, lhs, compiler.collection_name, rhs
        )
        if (
            (isinstance(lhs, ForeignKey) or isinstance(rhs, ForeignKey))
            and lhs_prepared.is_simple_column
            and rhs_prepared.is_simple_column
        ):
            # The join can be made using localField and foreignField.
            local_field = lhs_prepared.as_mql(compiler, connection)
            foreign_field = rhs_prepared.as_mql(compiler, connection)
        else:
            lhs_fields.append(lhs_prepared.as_mql(compiler, connection, as_expr=True))
            # In the lookup stage, the reference to this column doesn't include
            # the collection name.
            rhs_fields.append(rhs_prepared.as_mql(compiler, connection, as_expr=True))
    # Handle any join conditions besides matching field pairs.
    extra = self.join_field.get_extra_restriction(self.table_alias, self.parent_alias)
    extra_conditions = []
    if extra:
        replacements = _get_reroot_replacements(extra)
        extra_conditions.append(
            extra.replace_expressions(replacements).as_mql(compiler, connection)
        )
    # pushed_filter_expression is a Where expression from the outer WHERE
    # clause that involves fields from the joined (right-hand) table and
    # possibly the outer (left-hand) table. If it can be safely evaluated
    # within the $lookup pipeline (e.g., field comparisons like
    # right.status = left.id), it is "pushed" into the join's $match stage to
    # reduce the volume of joined documents. This only applies to INNER JOINs,
    # as pushing filters into a LEFT JOIN can change the semantics of the
    # result. LEFT JOINs may rely on null checks to detect missing RHS.
    if pushed_filter_expression and self.join_type == INNER:
        rerooted_replacement = _get_reroot_replacements(pushed_filter_expression)
        extra_conditions.append(
            pushed_filter_expression.replace_expressions(rerooted_replacement).as_mql(
                compiler, connection
            )
        )
    # Match the conditions:
    #   self.table_name.field1 = parent_table.field1
    # AND
    #   self.table_name.field2 = parent_table.field2
    # AND
    #   ...
    all_conditions = []
    if rhs_fields:
        all_conditions.append(
            {
                "$expr": {
                    "$and": [
                        {"$eq": [f"$${parent_template}{i}", field]}
                        for i, field in enumerate(rhs_fields)
                    ]
                }
            }
        )
    if extra_conditions:
        all_conditions.extend(extra_conditions)
    # Build matching pipeline
    num_conditions = len(all_conditions)
    if num_conditions == 0:
        pipeline = []
    elif num_conditions == 1:
        pipeline = [{"$match": all_conditions[0]}]
    else:
        pipeline = [{"$match": {"$and": all_conditions}}]
    lookup = {
        # The right-hand table to join.
        "from": self.table_name,
        "pipeline": pipeline,
        # Rename the output as table_alias.
        "as": self.table_alias,
    }
    if local_field and foreign_field:
        lookup.update(
            {
                "localField": local_field,
                "foreignField": foreign_field,
            }
        )
    if lhs_fields:
        lookup["let"] = {
            f"{parent_template}{i}": parent_field for i, parent_field in enumerate(lhs_fields)
        }
    lookup_pipeline = [{"$lookup": lookup}]
    # To avoid missing data when using $unwind, an empty collection is added if
    # the join isn't an inner join. For inner joins, rows with empty arrays are
    # removed, as $unwind unrolls or unnests the array and removes the row if
    # it's empty. This is the expected behavior for inner joins. For left outer
    # joins (LOUTER), however, an empty collection is returned.
    if self.join_type != INNER:
        lookup_pipeline.append(
            {
                "$set": {
                    self.table_alias: {
                        "$cond": {
                            "if": {
                                "$or": [
                                    {"$eq": [{"$type": f"${self.table_alias}"}, "missing"]},
                                    {"$eq": [{"$size": f"${self.table_alias}"}, 0]},
                                ]
                            },
                            "then": [{}],
                            "else": f"${self.table_alias}",
                        }
                    }
                }
            }
        )
    lookup_pipeline.append({"$unwind": f"${self.table_alias}"})
    return lookup_pipeline


def where_node(self, compiler, connection, as_expr=False):
    if self.connector == AND:
        full_needed, empty_needed = len(self.children), 1
    else:
        full_needed, empty_needed = 1, len(self.children)

    if self.connector == AND:
        operator = "$and"
    elif self.connector == XOR:
        # MongoDB doesn't support $xor, so convert:
        #   a XOR b XOR c XOR ...
        # to:
        #   (a OR b OR c OR ...) AND MOD(a + b + c + ..., 2) == 1
        # The result of an n-ary XOR is true when an odd number of operands
        # are true.
        lhs = self.__class__(self.children, OR)
        rhs_sum = reduce(
            add_operator,
            (Case(When(c, then=1), default=0) for c in self.children),
        )
        if len(self.children) > 2:
            rhs_sum = Mod(rhs_sum, 2)
        rhs = Exact(1, rhs_sum)
        return self.__class__([lhs, rhs], AND, self.negated).as_mql(
            compiler, connection, as_expr=as_expr
        )
    else:
        operator = "$or"

    children_mql = []
    for child in self.children:
        try:
            mql = child.as_mql(compiler, connection, as_expr=as_expr)
        except EmptyResultSet:
            empty_needed -= 1
        except FullResultSet:
            full_needed -= 1
        else:
            if mql:
                children_mql.append(mql)
            else:
                full_needed -= 1

        if empty_needed == 0:
            raise (FullResultSet if self.negated else EmptyResultSet)
        if full_needed == 0:
            raise (EmptyResultSet if self.negated else FullResultSet)

    if len(children_mql) == 1:
        mql = children_mql[0]
    elif len(children_mql) > 1:
        mql = {operator: children_mql}
    else:
        mql = {}

    if not mql:
        raise FullResultSet

    if self.negated and mql:
        mql = {"$not": [mql]} if as_expr else {"$nor": [mql]}

    return mql


def nothing_node(self, compiler, connection, as_expr=False):  # noqa: ARG001
    return self.as_sql(compiler, connection)


def register_nodes():
    ExtraWhere.as_mql = extra_where
    Join.as_mql = join
    NothingNode.as_mql = nothing_node
    WhereNode.as_mql = where_node



from django.db import models


class Reporter(models.Model):
    first_name = models.CharField(max_length=30)
    last_name = models.CharField(max_length=30)
    email = models.EmailField()

    class Meta:
        ordering = ("first_name", "last_name")

    def __str__(self):
        return f"{self.first_name} {self.last_name}".strip()



from collections import defaultdict

from django.core import checks
from django.core.exceptions import FieldDoesNotExist
from django.db.models import UniqueConstraint
from pymongo import ASCENDING
from pymongo.operations import IndexModel

from .fields import EmbeddedModelArrayField, PolymorphicEmbeddedModelArrayField
from .indexes import EmbeddedFieldIndexMixin, _get_condition_mql, get_field


def get_pymongo_index_model(self, model, schema_editor, field=None, column_prefix=""):
    """Return a pymongo IndexModel for this UniqueConstraint."""
    if self.contains_expressions:
        return None
    kwargs = {}
    filter_expression = defaultdict(dict)
    if self.condition:
        filter_expression.update(self._get_condition_mql(model, schema_editor))
    if self.nulls_distinct is None or self.nulls_distinct:
        # Indexing on $type matches the value of most SQL databases by allowing
        # multiple null values for the unique constraint. nulls_distinct will
        # be True or False for a UniqueConstraint, or None for
        # Field(unique=True) or Meta.unique_together.
        if field:
            column = column_prefix + field.column
            filter_expression[column].update({"$type": field.db_type(schema_editor.connection)})
        else:
            for field_name in self.fields:
                field_ = get_field(model, field_name)
                filter_expression[field_.column].update(
                    {"$type": field_.field.db_type(schema_editor.connection)}
                )
    if filter_expression:
        kwargs["partialFilterExpression"] = filter_expression
    index_orders = (
        [(column_prefix + field.column, ASCENDING)]
        if field
        else [
            (column_prefix + get_field(model, field_name).column, ASCENDING)
            for field_name in self.fields
        ]
    )
    return IndexModel(index_orders, name=self.name, unique=True, **kwargs)


class EmbeddedFieldUniqueConstraint(EmbeddedFieldIndexMixin, UniqueConstraint):
    meta_option_name = "constraints"

    def check(self, model, connection):
        errors = super().check(model, connection)
        # Due to MongoDB limitation (https://jira.mongodb.org/browse/SERVER-17853),
        # EmbeddedFieldUniqueConstraint cannot reference subfields of
        # EmbeddedModelArrayField unless nulls_distinct=False (so that the
        # index doesn't need to be created with a partialFilterExpression).
        if self.nulls_distinct is not False:
            for field_name in self.fields:
                # Get the top-level field, removing paths to embedded fields.
                local_field_name, *_ = field_name.split(".")
                try:
                    field = model._meta.get_field(local_field_name)
                except FieldDoesNotExist:
                    continue
                if isinstance(field, (EmbeddedModelArrayField, PolymorphicEmbeddedModelArrayField)):
                    errors.append(
                        checks.Error(
                            f"EmbeddedFieldUniqueConstraint {self.name!r} must "
                            "have nulls_distinct=False since it references "
                            f"{field.__class__.__name__} '{local_field_name}'.",
                            obj=model,
                            id="mongodb.constraints.embedded_unique.E001",
                        )
                    )
        return errors


def register_constraints():
    UniqueConstraint.get_pymongo_index_model = get_pymongo_index_model
    UniqueConstraint._get_condition_mql = _get_condition_mql



from pathlib import Path
from unittest import TestCase

from bson import encode, json_util

from .base import PerformanceTest
from .models import ForeignKeyModel, SmallFlatModel, SmallFlatModelFk


class SmallFlatDocTest(PerformanceTest):
    """Parent class for small flat document tests."""

    dataset = "small_doc.json"

    def setUp(self):
        super().setUp()
        with open(  # noqa: PTH123
            Path(self.test_data_path) / Path("flat-models") / self.dataset
        ) as data:
            self.document = json_util.loads(data.read())
        self.setUpData()
        self.data_size = len(encode(self.document)) * self.num_docs

    def setUpData(self):
        SmallFlatModel.objects.bulk_create(
            SmallFlatModel(**self.document) for _ in range(self.num_docs)
        )

    def tearDown(self):
        super().tearDown()
        SmallFlatModel.objects.all().delete()


class TestSmallFlatDocCreation(SmallFlatDocTest, TestCase):
    """Creating a small flat document."""

    def setUpData(self):
        # Don't create data since this is the creation test.
        pass

    def do_task(self):
        for _ in range(self.num_docs):
            SmallFlatModel.objects.create(**self.document)

    def after(self):
        SmallFlatModel.objects.all().delete()


class TestSmallFlatDocUpdate(SmallFlatDocTest, TestCase):
    """Updating a field within a small flat document."""

    def setUp(self):
        super().setUp()
        self.models = list(SmallFlatModel.objects.all())
        self.data_size = len(encode({"field1": "updated_value0"})) * self.num_docs
        self.iteration = 0

    def do_task(self):
        for model in self.models:
            model.field1 = "updated_value" + str(self.iteration)
            model.save()
        self.iteration += 1


class TestSmallFlatDocFilterById(SmallFlatDocTest, TestCase):
    """Filtering small flat documents by their primary key."""

    def setUp(self):
        super().setUp()
        self.ids = SmallFlatModel.objects.values_list("id", flat=True)

    def do_task(self):
        for _id in self.ids:
            list(SmallFlatModel.objects.filter(id=_id))


class TestSmallFlatDocFilterByForeignKey(SmallFlatDocTest, TestCase):
    """Filtering small flat documents by a foreign key field."""

    def setUp(self):
        super().setUp()
        self.fks = []
        for _ in range(self.num_docs):
            foreign_key_model = ForeignKeyModel.objects.create(name="foreign_key_name")
            self.fks.append(foreign_key_model.id)
            model = SmallFlatModelFk(**self.document)
            model.field_fk = foreign_key_model
            model.save()

    def setUpData(self):
        # Don't create data since SmallFlatModel isn't used.
        pass

    def do_task(self):
        for fk in self.fks:
            list(SmallFlatModelFk.objects.filter(field_fk=fk))

    def tearDown(self):
        super().tearDown()
        SmallFlatModelFk.objects.all().delete()
        ForeignKeyModel.objects.all().delete()


class TestSmallFlatDocFilterPkByIn(SmallFlatDocTest, TestCase):
    """Filtering small flat documents using __in for primary keys."""

    def setUp(self):
        super().setUp()
        self.ids = SmallFlatModel.objects.values_list("id", flat=True)[:100]

    def do_task(self):
        list(SmallFlatModel.objects.filter(id__in=self.ids))



from django.contrib.gis.db import models


class NamedModel(models.Model):
    name = models.CharField(max_length=30)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name


class City(NamedModel):
    point = models.PointField()


class Zipcode(models.Model):
    code = models.CharField(max_length=10)
    poly = models.PolygonField(geography=True)



from pathlib import Path
from unittest import TestCase

from bson import ObjectId, encode, json_util

from .base import PerformanceTest
from .models import IntegerEmbeddedModel, LargeNestedModel, StringEmbeddedModel


class LargeNestedDocTest(PerformanceTest):
    """Parent class for large nested document tests."""

    dataset = "large_doc_nested.json"

    def setUp(self):
        super().setUp()
        with open(  # noqa: PTH123
            Path(self.test_data_path) / Path("nested-models") / self.dataset
        ) as data:
            self.document = json_util.loads(data.read())
        self.setUpData()
        self.data_size = len(encode(self.document)) * self.num_docs

    def setUpData(self):
        for _ in range(self.num_docs):
            model = LargeNestedModel()
            for field_name, model_data in self.document.items():
                if "array" in field_name:
                    array_models = []
                    for item in model_data:
                        embedded_str_model = StringEmbeddedModel(**item)
                        embedded_str_model.unique_field = str(ObjectId())
                        array_models.append(embedded_str_model)
                    setattr(model, field_name, array_models)
                elif "embedded_str_doc" in field_name:
                    embedded_str_model = StringEmbeddedModel(**model_data)
                    embedded_str_model.unique_field = str(ObjectId())
                    setattr(model, field_name, embedded_str_model)
                else:
                    embedded_int_model = IntegerEmbeddedModel(**model_data)
                    setattr(model, field_name, embedded_int_model)
            model.save()

    def tearDown(self):
        super().tearDown()
        LargeNestedModel.objects.all().delete()


class TestLargeNestedDocCreation(LargeNestedDocTest, TestCase):
    """Creating a large nested document."""

    def setUpData(self):
        # Don't create data since this is the creation test.
        pass

    def do_task(self):
        for _ in range(self.num_docs):
            model = LargeNestedModel()
            for field_name, model_data in self.document.items():
                if "array" in field_name:
                    array_models = []
                    for item in model_data:
                        embedded_str_model = StringEmbeddedModel(**item)
                        embedded_str_model.unique_field = str(ObjectId())
                        array_models.append(embedded_str_model)
                    setattr(model, field_name, array_models)
                elif "embedded_str_doc" in field_name:
                    embedded_str_model = StringEmbeddedModel(**model_data)
                    embedded_str_model.unique_field = str(ObjectId())
                    setattr(model, field_name, embedded_str_model)
                else:
                    embedded_int_model = IntegerEmbeddedModel(**model_data)
                    setattr(model, field_name, embedded_int_model)
            model.save()

    def after(self):
        LargeNestedModel.objects.all().delete()


class TestLargeNestedDocUpdate(LargeNestedDocTest, TestCase):
    """Updating an embedded field within a large nested document."""

    def setUp(self):
        super().setUp()
        self.models = list(LargeNestedModel.objects.all())
        self.data_size = len(encode({"field1": "updated_value0"})) * self.num_docs
        self.iteration = 0

    def do_task(self):
        for model in self.models:
            model.embedded_str_doc_1.field1 = "updated_value" + str(self.iteration)
            model.save()
        self.iteration += 1


class TestLargeNestedDocFilterByUniqueField(LargeNestedDocTest, TestCase):
    """
    Filtering large nested documents by a unique field in an embedded document.
    """

    def setUp(self):
        super().setUp()
        self.setUpData()
        self.ids = LargeNestedModel.objects.values_list(
            "embedded_str_doc_1__unique_field", flat=True
        )

    def do_task(self):
        for _id in self.ids:
            list(LargeNestedModel.objects.filter(embedded_str_doc_1__unique_field=_id))


class TestLargeNestedDocFilterArrayByUniqueField(LargeNestedDocTest, TestCase):
    """
    Filtering large nested documents by a unique field in an embedded document
    array.
    """

    def setUp(self):
        super().setUp()
        self.ids = LargeNestedModel.objects.values_list(
            "embedded_str_doc_array__0__unique_field", flat=True
        )

    def do_task(self):
        for _id in self.ids:
            list(LargeNestedModel.objects.filter(embedded_str_doc_array__unique_field=_id))



import datetime
from decimal import Decimal

from django.db.models import Count, Max, Q
from django.test import TestCase

from .models import Author, Book


class FilteredAggregateTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.a1 = Author.objects.create(name="test", age=40)
        cls.a2 = Author.objects.create(name="test2", age=60)
        cls.a3 = Author.objects.create(name="test3", age=40)
        cls.b1 = Book.objects.create(
            isbn="159059725",
            name="The Definitive Guide to Django: Web Development Done Right",
            pages=447,
            rating=4.5,
            price=Decimal("30.00"),
            contact=cls.a1,
            pubdate=datetime.date(2007, 12, 6),
        )
        cls.b2 = Book.objects.create(
            isbn="067232959",
            name="Sams Teach Yourself Django in 24 Hours",
            pages=528,
            rating=3.0,
            price=Decimal("30.00"),
            contact=cls.a2,
            pubdate=datetime.date(2008, 3, 3),
        )
        cls.b3 = Book.objects.create(
            isbn="159059996",
            name="Practical Django Projects",
            pages=600,
            rating=40.5,
            price=Decimal("30.00"),
            contact=cls.a3,
            pubdate=datetime.date(2008, 6, 23),
        )
        cls.a1.friends.add(cls.a2)
        cls.a1.friends.add(cls.a3)
        cls.b1.authors.add(cls.a1)
        cls.b1.authors.add(cls.a3)
        cls.b2.authors.add(cls.a2)
        cls.b3.authors.add(cls.a3)

    def test_filtered_aggregate_empty_condition_distinct(self):
        book = Book.objects.annotate(
            ages=Count("authors__age", filter=Q(authors__in=[]), distinct=True),
        ).get(pk=self.b1.pk)
        self.assertEqual(book.ages, 0)
        aggregate = Book.objects.aggregate(max_rating=Max("rating", filter=Q(rating__in=[])))
        self.assertEqual(aggregate, {"max_rating": None})

    def test_filtered_aggregate_full_condition(self):
        book = Book.objects.annotate(
            ages=Count("authors__age", filter=~Q(pk__in=[])),
        ).get(pk=self.b1.pk)
        self.assertEqual(book.ages, 2)
        aggregate = Book.objects.aggregate(max_rating=Max("rating", filter=~Q(rating__in=[])))
        self.assertEqual(aggregate, {"max_rating": 40.5})

    def test_filtered_aggregate_full_condition_distinct(self):
        book = Book.objects.annotate(
            ages=Count("authors__age", filter=~Q(authors__in=[]), distinct=True),
        ).get(pk=self.b1.pk)
        self.assertEqual(book.ages, 1)
        aggregate = Book.objects.aggregate(max_rating=Max("rating", filter=~Q(rating__in=[])))
        self.assertEqual(aggregate, {"max_rating": 40.5})



from django.db import connection


class SchemaAssertionMixin:
    def assertAddRemoveIndex(self, editor, model, index):
        editor.add_index(index=index, model=model)
        try:
            self.assertIn(
                index.name,
                connection.introspection.get_constraints(
                    cursor=None,
                    table_name=model._meta.db_table,
                ),
            )
        finally:
            editor.remove_index(index=index, model=model)
        self.assertNotIn(
            index.name,
            connection.introspection.get_constraints(
                cursor=None,
                table_name=model._meta.db_table,
            ),
        )



# Django settings for the performance tests.
DATABASES = {
    "default": {
        "ENGINE": "django_mongodb_backend",
        "NAME": "django-performance-tests",
    },
}

SECRET_KEY = "django_tests_secret_key"  # noqa: S105

INSTALLED_APPS = ["perf"]

DEFAULT_AUTO_FIELD = "django_mongodb_backend.fields.ObjectIdAutoField"

TEST_RUNNER = "perf.runner.TestRunner"



from django.test.runner import DiscoverRunner

from .base import write_output_file


class TestRunner(DiscoverRunner):
    def teardown_test_environment(self, **kwargs):
        write_output_file()
        super().teardown_test_environment(**kwargs)



from django.db import models

from django_mongodb_backend.fields import EmbeddedModelArrayField, EmbeddedModelField
from django_mongodb_backend.models import EmbeddedModel


class SmallFlatModel(models.Model):
    field1 = models.CharField(max_length=100)
    field2 = models.CharField(max_length=100)
    field3 = models.CharField(max_length=100)
    field4 = models.CharField(max_length=100)
    field5 = models.CharField(max_length=100)
    field6 = models.CharField(max_length=100)
    field7 = models.CharField(max_length=100)
    field8 = models.IntegerField()
    field9 = models.IntegerField()
    field10 = models.IntegerField()
    field11 = models.IntegerField()
    field12 = models.IntegerField()
    field13 = models.IntegerField()

    class Meta:
        indexes = [
            models.Index(fields=["field1"], name="field1_idx"),
        ]


class ForeignKeyModel(models.Model):
    name = models.CharField(max_length=100)


class SmallFlatModelFk(models.Model):
    field1 = models.CharField(max_length=100)
    field2 = models.CharField(max_length=100)
    field3 = models.CharField(max_length=100)
    field4 = models.CharField(max_length=100)
    field5 = models.CharField(max_length=100)
    field6 = models.CharField(max_length=100)
    field7 = models.CharField(max_length=100)
    field8 = models.IntegerField()
    field9 = models.IntegerField()
    field10 = models.IntegerField()
    field11 = models.IntegerField()
    field12 = models.IntegerField()
    field13 = models.IntegerField()
    field_fk = models.ForeignKey(ForeignKeyModel, on_delete=models.DO_NOTHING)


# Construct LargeFlatModel programmatically to avoid a very long model
# definition.
large_flat_model_attrs = {
    "__module__": "perf.models",
    "image_field": models.BinaryField(),
}
for i in range(125):
    large_flat_model_attrs[f"field{i + 1}"] = models.CharField(max_length=100)
    large_flat_model_attrs[f"field{i + 126}"] = models.IntegerField()

LargeFlatModel = type("LargeFlatModel", (models.Model,), large_flat_model_attrs)


class StringEmbeddedModel(EmbeddedModel):
    unique_field = models.CharField(max_length=100)
    field1 = models.CharField(max_length=100)
    field2 = models.CharField(max_length=100)
    field3 = models.CharField(max_length=100)
    field4 = models.CharField(max_length=100)
    field5 = models.CharField(max_length=100)
    field6 = models.CharField(max_length=100)
    field7 = models.CharField(max_length=100)
    field8 = models.CharField(max_length=100)
    field9 = models.CharField(max_length=100)
    field10 = models.CharField(max_length=100)
    field11 = models.CharField(max_length=100)
    field12 = models.CharField(max_length=100)
    field13 = models.CharField(max_length=100)
    field14 = models.CharField(max_length=100)
    field15 = models.CharField(max_length=100)


class IntegerEmbeddedModel(EmbeddedModel):
    field1 = models.IntegerField()
    field2 = models.IntegerField()
    field3 = models.IntegerField()
    field4 = models.IntegerField()
    field5 = models.IntegerField()
    field6 = models.IntegerField()
    field7 = models.IntegerField()
    field8 = models.IntegerField()
    field9 = models.IntegerField()
    field10 = models.IntegerField()
    field11 = models.IntegerField()
    field12 = models.IntegerField()
    field13 = models.IntegerField()
    field14 = models.IntegerField()
    field15 = models.IntegerField()


class LargeNestedModel(models.Model):
    embedded_str_doc_1 = EmbeddedModelField(StringEmbeddedModel)
    embedded_str_doc_2 = EmbeddedModelField(StringEmbeddedModel)
    embedded_str_doc_3 = EmbeddedModelField(StringEmbeddedModel)
    embedded_str_doc_4 = EmbeddedModelField(StringEmbeddedModel)
    embedded_str_doc_5 = EmbeddedModelField(StringEmbeddedModel)
    embedded_str_doc_array = EmbeddedModelArrayField(StringEmbeddedModel)
    embedded_int_doc_8 = EmbeddedModelField(IntegerEmbeddedModel)
    embedded_int_doc_9 = EmbeddedModelField(IntegerEmbeddedModel)
    embedded_int_doc_10 = EmbeddedModelField(IntegerEmbeddedModel)
    embedded_int_doc_11 = EmbeddedModelField(IntegerEmbeddedModel)
    embedded_int_doc_12 = EmbeddedModelField(IntegerEmbeddedModel)
    embedded_int_doc_13 = EmbeddedModelField(IntegerEmbeddedModel)
    embedded_int_doc_14 = EmbeddedModelField(IntegerEmbeddedModel)



from pymongo import GEOSPHERE
from pymongo.operations import IndexModel


class GISSchemaEditor:
    def _create_model_indexes(self, model, column_prefix="", parent_model=None):
        super()._create_model_indexes(model, column_prefix, parent_model)
        for field in model._meta.local_fields:
            if getattr(field, "spatial_index", False):
                self._add_spatial_index(parent_model or model, field, column_prefix)

    def add_field(self, model, field):
        super().add_field(model, field)
        if getattr(field, "spatial_index", False):
            self._add_spatial_index(model, field)

    def _alter_field(
        self,
        model,
        old_field,
        new_field,
        old_type,
        new_type,
        old_db_params,
        new_db_params,
        strict=False,
    ):
        super()._alter_field(
            model,
            old_field,
            new_field,
            old_type,
            new_type,
            old_db_params,
            new_db_params,
            strict=strict,
        )
        old_field_spatial_index = getattr(old_field, "spatial_index", False)
        new_field_spatial_index = getattr(new_field, "spatial_index", False)
        if not old_field_spatial_index and new_field_spatial_index:
            self._add_spatial_index(model, new_field)
        elif old_field_spatial_index and not new_field_spatial_index:
            self._delete_spatial_index(model, new_field)

    def remove_field(self, model, field):
        super().remove_field(model, field)
        if getattr(field, "spatial_index", False):
            self._delete_spatial_index(model, field)

    def _add_spatial_index(self, model, field, column_prefix=""):
        index_name = self._create_spatial_index_name(model, field, column_prefix)
        self.get_collection(model._meta.db_table).create_indexes(
            [IndexModel([(column_prefix + field.column, GEOSPHERE)], name=index_name)]
        )

    def _delete_spatial_index(self, model, field):
        index_name = self._create_spatial_index_name(model, field)
        self.get_collection(model._meta.db_table).drop_index(index_name)

    def _create_spatial_index_name(self, model, field, column_prefix=""):
        return f"{model._meta.db_table}_{column_prefix}{field.column}_id"



from .array import ArrayField
from .auto import ObjectIdAutoField
from .duration import register_duration_field
from .embedded_model import EmbeddedModelField
from .embedded_model_array import EmbeddedModelArrayField
from .encryption import (
    EncryptedArrayField,
    EncryptedBigIntegerField,
    EncryptedBinaryField,
    EncryptedBooleanField,
    EncryptedCharField,
    EncryptedDateField,
    EncryptedDateTimeField,
    EncryptedDecimalField,
    EncryptedDurationField,
    EncryptedEmailField,
    EncryptedEmbeddedModelArrayField,
    EncryptedEmbeddedModelField,
    EncryptedFieldMixin,
    EncryptedFloatField,
    EncryptedGenericIPAddressField,
    EncryptedIntegerField,
    EncryptedObjectIdField,
    EncryptedPositiveBigIntegerField,
    EncryptedPositiveIntegerField,
    EncryptedPositiveSmallIntegerField,
    EncryptedSmallIntegerField,
    EncryptedTextField,
    EncryptedTimeField,
    EncryptedURLField,
    EncryptedUUIDField,
)
from .json import register_json_field
from .objectid import ObjectIdField
from .polymorphic_embedded_model import PolymorphicEmbeddedModelField
from .polymorphic_embedded_model_array import PolymorphicEmbeddedModelArrayField

__all__ = [
    "ArrayField",
    "EmbeddedModelArrayField",
    "EmbeddedModelField",
    "EncryptedArrayField",
    "EncryptedBigIntegerField",
    "EncryptedBinaryField",
    "EncryptedBooleanField",
    "EncryptedCharField",
    "EncryptedDateField",
    "EncryptedDateTimeField",
    "EncryptedDecimalField",
    "EncryptedDurationField",
    "EncryptedEmailField",
    "EncryptedEmbeddedModelArrayField",
    "EncryptedEmbeddedModelField",
    "EncryptedFieldMixin",
    "EncryptedFloatField",
    "EncryptedGenericIPAddressField",
    "EncryptedIntegerField",
    "EncryptedObjectIdField",
    "EncryptedPositiveBigIntegerField",
    "EncryptedPositiveIntegerField",
    "EncryptedPositiveSmallIntegerField",
    "EncryptedSmallIntegerField",
    "EncryptedTextField",
    "EncryptedTimeField",
    "EncryptedURLField",
    "EncryptedUUIDField",
    "ObjectIdAutoField",
    "ObjectIdField",
    "PolymorphicEmbeddedModelArrayField",
    "PolymorphicEmbeddedModelField",
    "register_fields",
]


def register_fields():
    register_duration_field()
    register_json_field()



from django import forms
from django.core.exceptions import ValidationError
from django.forms import formset_factory, model_to_dict
from django.forms.models import modelform_factory
from django.utils.html import format_html, format_html_join


class EmbeddedModelArrayField(forms.Field):
    def __init__(self, model, *, prefix, max_num=None, extra_forms=3, **kwargs):
        self.model = model
        self.prefix = prefix
        self.formset = formset_factory(
            form=modelform_factory(model, fields="__all__"),
            can_delete=True,
            max_num=max_num,
            extra=extra_forms,
            validate_max=True,
        )
        kwargs["widget"] = EmbeddedModelArrayWidget()
        super().__init__(**kwargs)

    def clean(self, value):
        if not value:
            return []
        formset = self.formset(value, prefix=self.prefix_override or self.prefix)
        if not formset.is_valid():
            raise ValidationError(formset.errors + formset.non_form_errors())
        cleaned_data = []
        for data in formset.cleaned_data:
            # The "delete" checkbox isn't part of model data and must be
            # removed. The fallback to True skips empty forms.
            if data.pop("DELETE", True):
                continue
            cleaned_data.append(self.model(**data))
        return cleaned_data

    def has_changed(self, initial, data):
        formset = self.formset(data, initial=models_to_dicts(initial), prefix=self.prefix)
        return formset.has_changed()

    def get_bound_field(self, form, field_name):
        # Nested embedded model form fields need a double prefix.
        # HACK: Setting self.prefix_override makes it available in clean()
        # which doesn't have access to the form.
        self.prefix_override = f"{form.prefix}-{self.prefix}" if form.prefix else None
        return EmbeddedModelArrayBoundField(form, self, field_name, self.prefix_override)


class EmbeddedModelArrayBoundField(forms.BoundField):
    def __init__(self, form, field, name, prefix_override):
        super().__init__(form, field, name)
        self.formset = field.formset(
            self.data if form.is_bound else None,
            initial=models_to_dicts(self.initial),
            prefix=prefix_override if prefix_override else self.html_name,
        )

    def __str__(self):
        body = format_html_join(
            "\n", "<tbody>{}</tbody>", ((form.as_table(),) for form in self.formset)
        )
        return format_html("<table>\n{}\n</table>\n{}", body, self.formset.management_form)


class EmbeddedModelArrayWidget(forms.Widget):
    """
    Extract the data for EmbeddedModelArrayFormField's formset.
    This widget is never rendered.
    """

    def value_from_datadict(self, data, files, name):
        return {field: value for field, value in data.items() if field.startswith(f"{name}-")}


def models_to_dicts(models):
    """
    Convert initial data (which is a list of model instances or None) to a
    list of dictionary data suitable for a formset.
    """
    return [model_to_dict(model) for model in models or []]



import json

from django.core import checks, exceptions
from django.db.models import Field, Func, IntegerField, Transform, Value
from django.db.models.fields.mixins import CheckFieldDefaultMixin
from django.db.models.lookups import Exact, FieldGetDbPrepValueMixin, In, Lookup
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from django_mongodb_backend.forms import SimpleArrayField
from django_mongodb_backend.query_utils import is_constant_value, process_lhs, process_rhs
from django_mongodb_backend.utils import prefix_validation_error
from django_mongodb_backend.validators import ArrayMaxLengthValidator, LengthValidator

__all__ = ["ArrayField"]


class AttributeSetter:
    def __init__(self, name, value):
        setattr(self, name, value)


class ArrayField(CheckFieldDefaultMixin, Field):
    empty_strings_allowed = False
    default_error_messages = {
        "item_invalid": _("Item %(nth)s in the array did not validate:"),
        "nested_array_mismatch": _("Nested arrays must have the same length."),
    }
    _default_hint = ("list", "[]")

    def __init__(self, base_field, max_size=None, size=None, **kwargs):
        self.base_field = base_field
        self.max_size = max_size
        self.size = size
        if self.max_size:
            self.default_validators = [
                *self.default_validators,
                ArrayMaxLengthValidator(self.max_size),
            ]
        if self.size:
            self.default_validators = [
                *self.default_validators,
                LengthValidator(self.size),
            ]
        # For performance, only add a from_db_value() method if the base field
        # implements it.
        if hasattr(self.base_field, "from_db_value"):
            self.from_db_value = self._from_db_value
        super().__init__(**kwargs)

    @property
    def model(self):
        try:
            return self.__dict__["model"]
        except KeyError:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute 'model'"
            ) from None

    @model.setter
    def model(self, model):
        self.__dict__["model"] = model
        self.base_field.model = model

    @classmethod
    def _choices_is_value(cls, value):
        return isinstance(value, (list, tuple)) or super()._choices_is_value(value)

    def check(self, **kwargs):
        errors = super().check(**kwargs)
        if self.base_field.remote_field:
            errors.append(
                checks.Error(
                    "Base field for array cannot be a related field.",
                    obj=self,
                    id="mongodb.fields.array.E002",
                )
            )
        else:
            base_checks = self.base_field.check()
            if base_checks:
                error_messages = "\n    ".join(
                    f"{base_check.msg} ({base_check.id})"
                    for base_check in base_checks
                    if isinstance(base_check, checks.Error)
                )
                if error_messages:
                    errors.append(
                        checks.Error(
                            f"Base field for array has errors:\n    {error_messages}",
                            obj=self,
                            id="mongodb.fields.array.E001",
                        )
                    )
                warning_messages = "\n    ".join(
                    f"{base_check.msg} ({base_check.id})"
                    for base_check in base_checks
                    if isinstance(base_check, checks.Warning)
                )
                if warning_messages:
                    errors.append(
                        checks.Warning(
                            f"Base field for array has warnings:\n    {warning_messages}",
                            obj=self,
                            id="mongodb.fields.array.W004",
                        )
                    )
        if self.size and self.max_size:
            errors.append(
                checks.Error(
                    "ArrayField cannot have both size and max_size.",
                    obj=self,
                    id="mongodb.fields.array.E003",
                )
            )
        return errors

    def set_attributes_from_name(self, name):
        super().set_attributes_from_name(name)
        self.base_field.set_attributes_from_name(name)

    @property
    def description(self):
        return f"Array of {self.base_field.description}"

    def db_type(self, connection):
        return "array"

    def get_db_prep_value(self, value, connection, prepared=False):
        if isinstance(value, (list, tuple)):
            return [self.base_field.get_db_prep_value(i, connection, prepared=False) for i in value]
        return value

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if path == "django_mongodb_backend.fields.array.ArrayField":
            path = "django_mongodb_backend.fields.ArrayField"
        kwargs["base_field"] = self.base_field.clone()
        if self.max_size is not None:
            kwargs["max_size"] = self.max_size
        if self.size is not None:
            kwargs["size"] = self.size
        return name, path, args, kwargs

    def to_python(self, value):
        if isinstance(value, str):
            # Assume value is being deserialized.
            vals = json.loads(value)
            value = [self.base_field.to_python(val) for val in vals]
        return value

    def _from_db_value(self, value, expression, connection):
        if value is None:
            return value
        return [self.base_field.from_db_value(item, expression, connection) for item in value]

    def value_to_string(self, obj):
        values = []
        vals = self.value_from_object(obj)
        base_field = self.base_field

        for val in vals:
            if val is None:
                values.append(None)
            else:
                obj = AttributeSetter(base_field.attname, val)
                values.append(base_field.value_to_string(obj))
        return json.dumps(values)

    def get_transform(self, name):
        transform = super().get_transform(name)
        if transform:
            return transform
        if "_" not in name:
            try:
                index = int(name)
            except ValueError:
                pass
            else:
                return IndexTransformFactory(index, self.base_field)
        try:
            start, end = name.split("_")
            start = int(start)
            end = int(end)
        except ValueError:
            pass
        else:
            return SliceTransformFactory(start, end)

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        for index, part in enumerate(value):
            try:
                self.base_field.validate(part, model_instance)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["item_invalid"],
                    code="item_invalid",
                    params={"nth": index + 1},
                ) from None
        if isinstance(self.base_field, ArrayField) and len({len(i) for i in value}) > 1:
            raise exceptions.ValidationError(
                self.error_messages["nested_array_mismatch"],
                code="nested_array_mismatch",
            )

    def run_validators(self, value):
        super().run_validators(value)
        for index, part in enumerate(value):
            try:
                self.base_field.run_validators(part)
            except exceptions.ValidationError as error:
                raise prefix_validation_error(
                    error,
                    prefix=self.error_messages["item_invalid"],
                    code="item_invalid",
                    params={"nth": index + 1},
                ) from None

    def formfield(self, **kwargs):
        return super().formfield(
            **{
                "form_class": SimpleArrayField,
                "base_field": self.base_field.formfield(),
                "max_length": self.max_size,
                "length": self.size,
                **kwargs,
            }
        )


class Array(Func):
    def as_mql_expr(self, compiler, connection):
        return [
            expr.as_mql(compiler, connection, as_expr=True)
            for expr in self.get_source_expressions()
        ]

    def as_mql_path(self, compiler, connection):
        return [expr.as_mql(compiler, connection) for expr in self.get_source_expressions()]

    @cached_property
    def can_use_path(self):
        return all(is_constant_value(expr) for expr in self.get_source_expressions())

    @property
    def is_simple_column(self):
        return False


class ArrayRHSMixin:
    def __init__(self, lhs, rhs):
        if isinstance(rhs, (tuple, list)):
            expressions = []
            for value in rhs:
                if not hasattr(value, "resolve_expression"):
                    field = lhs.output_field
                    value = Value(field.base_field.get_prep_value(value))
                expressions.append(value)
            rhs = Array(*expressions)
        super().__init__(lhs, rhs)


@ArrayField.register_lookup
class ArrayContains(ArrayRHSMixin, FieldGetDbPrepValueMixin, Lookup):
    lookup_name = "contains"

    def as_mql_expr(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        value = process_rhs(self, compiler, connection, as_expr=True)
        return {
            "$and": [
                {"$ne": [lhs_mql, None]},
                {"$ne": [value, None]},
                {"$setIsSubset": [value, lhs_mql]},
            ]
        }

    def as_mql_path(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        value = process_rhs(self, compiler, connection)
        return {lhs_mql: {"$all": value}}


@ArrayField.register_lookup
class ArrayContainedBy(ArrayRHSMixin, FieldGetDbPrepValueMixin, Lookup):
    lookup_name = "contained_by"

    def as_mql_expr(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        value = process_rhs(self, compiler, connection, as_expr=True)
        return {
            "$and": [
                {"$ne": [lhs_mql, None]},
                {"$ne": [value, None]},
                {"$setIsSubset": [lhs_mql, value]},
            ]
        }


@ArrayField.register_lookup
class ArrayExact(ArrayRHSMixin, Exact):
    pass


@ArrayField.register_lookup
class ArrayOverlap(ArrayRHSMixin, FieldGetDbPrepValueMixin, Lookup):
    lookup_name = "overlap"

    def get_subquery_wrapping_pipeline(self, compiler, connection, field_name, expr):
        return [
            {"$project": {"subquery_results": expr.as_mql(compiler, connection, as_expr=True)}},
            {"$unwind": "$subquery_results"},
            {
                "$group": {
                    "_id": None,
                    "subquery_results": {"$addToSet": "$subquery_results"},
                }
            },
            # Workaround for https://jira.mongodb.org/browse/SERVER-114196:
            # $$NOW becomes unavailable after $unionWith, so it must be stored
            # beforehand to ensure it remains accessible later in the pipeline.
            {"$addFields": {"__now": "$$NOW"}},
            # Add an extra empty document to handle default values on empty
            # results.
            {"$unionWith": {"pipeline": [{"$documents": [{"subquery_results": []}]}]}},
            {"$limit": 1},
            {"$project": {field_name: "$subquery_results"}},
        ]

    def as_mql_expr(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        value = process_rhs(self, compiler, connection, as_expr=True)
        return {
            "$and": [
                {"$ne": [lhs_mql, None]},
                {"$size": {"$setIntersection": [value, lhs_mql]}},
            ]
        }

    def as_mql_path(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        value = process_rhs(self, compiler, connection)
        return {lhs_mql: {"$in": value}}


@ArrayField.register_lookup
class ArrayLenTransform(Transform):
    lookup_name = "len"
    output_field = IntegerField()

    def as_mql_expr(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        return {"$cond": {"if": {"$isArray": lhs_mql}, "then": {"$size": lhs_mql}, "else": None}}


@ArrayField.register_lookup
class ArrayInLookup(In):
    def get_prep_lookup(self):
        values = super().get_prep_lookup()
        if hasattr(values, "resolve_expression"):
            return values
        # process_rhs() expects hashable values, so convert lists to tuples.
        prepared_values = []
        for value in values:
            if hasattr(value, "resolve_expression"):
                prepared_values.append(value)
            else:
                prepared_values.append(tuple(value))
        return prepared_values


class IndexTransform(Transform):
    def __init__(self, index, base_field, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.index = index
        self.base_field = base_field

    def as_mql_expr(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        return {"$arrayElemAt": [lhs_mql, self.index]}

    def as_mql_path(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        return f"{lhs_mql}.{self.index}"

    @property
    def output_field(self):
        return self.base_field

    @property
    def can_use_path(self):
        return self.is_simple_column

    @property
    def is_simple_column(self):
        return self.lhs.is_simple_column


class IndexTransformFactory:
    def __init__(self, index, base_field):
        self.index = index
        self.base_field = base_field

    def __call__(self, *args, **kwargs):
        return IndexTransform(self.index, self.base_field, *args, **kwargs)


class SliceTransform(Transform):
    def __init__(self, start, end, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start = start
        self.end = end

    def as_mql_expr(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection, as_expr=True)
        return {"$slice": [lhs_mql, self.start, self.end]}


class SliceTransformFactory:
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __call__(self, *args, **kwargs):
        return SliceTransform(self.start, self.end, *args, **kwargs)



from bson import ObjectId
from django.urls import register_converter


class ObjectIdConverter:
    regex = "[a-f0-9]{24}"

    def to_python(self, value):
        return ObjectId(value)

    def to_url(self, value):
        return str(value)


def register_urls():
    register_converter(ObjectIdConverter, "object_id")



from django.db import models

from django_mongodb_backend.fields import (
    EmbeddedModelField,
    EncryptedArrayField,
    EncryptedBigIntegerField,
    EncryptedBinaryField,
    EncryptedBooleanField,
    EncryptedCharField,
    EncryptedDateField,
    EncryptedDateTimeField,
    EncryptedDecimalField,
    EncryptedDurationField,
    EncryptedEmailField,
    EncryptedEmbeddedModelArrayField,
    EncryptedEmbeddedModelField,
    EncryptedFloatField,
    EncryptedGenericIPAddressField,
    EncryptedIntegerField,
    EncryptedObjectIdField,
    EncryptedPositiveBigIntegerField,
    EncryptedPositiveIntegerField,
    EncryptedPositiveSmallIntegerField,
    EncryptedSmallIntegerField,
    EncryptedTextField,
    EncryptedTimeField,
    EncryptedURLField,
    EncryptedUUIDField,
)
from django_mongodb_backend.models import EmbeddedModel


class Author(models.Model):
    name = models.CharField(max_length=255)


class Book(models.Model):
    title = models.CharField(max_length=255)
    author = models.ForeignKey(Author, models.CASCADE)


class EncryptedModel(models.Model):
    class Meta:
        abstract = True
        required_db_features = {"supports_queryable_encryption"}


# Array models
class ArrayModel(EncryptedModel):
    values = EncryptedArrayField(models.IntegerField(), size=5)


# Embedded models
class Patient(EncryptedModel):
    patient_name = models.CharField(max_length=255)
    patient_id = models.BigIntegerField()
    patient_record = EmbeddedModelField("PatientRecord")

    def __str__(self):
        return f"{self.patient_name} ({self.patient_id})"


class PatientRecord(EmbeddedModel):
    ssn = EncryptedCharField(max_length=11, queries={"queryType": "equality"})
    billing = EncryptedEmbeddedModelField("Billing")
    bill_amount = models.DecimalField(max_digits=10, decimal_places=2)


class Billing(EmbeddedModel):
    cc_type = models.CharField(max_length=50)
    cc_number = models.CharField(max_length=20)


# Embedded array models
class Actor(EmbeddedModel):
    name = models.CharField(max_length=100)


class Movie(EncryptedModel):
    title = models.CharField(max_length=200)
    cast = EncryptedEmbeddedModelArrayField(Actor)

    def __str__(self):
        return self.title


# Equality-queryable models
class BinaryModel(EncryptedModel):
    value = EncryptedBinaryField(queries={"queryType": "equality"})


class BooleanModel(EncryptedModel):
    value = EncryptedBooleanField(queries={"queryType": "equality"})


class CharModel(EncryptedModel):
    value = EncryptedCharField(max_length=255, queries={"queryType": "equality"})
    plain = models.CharField(max_length=255, blank=True)


class EmailModel(EncryptedModel):
    value = EncryptedEmailField(max_length=255, queries={"queryType": "equality"})


class GenericIPAddressModel(EncryptedModel):
    value = EncryptedGenericIPAddressField(queries={"queryType": "equality"})


class ObjectIdModel(EncryptedModel):
    value = EncryptedObjectIdField(queries={"queryType": "equality"})


class TextModel(EncryptedModel):
    value = EncryptedTextField(queries={"queryType": "equality"})


class URLModel(EncryptedModel):
    value = EncryptedURLField(max_length=500, queries={"queryType": "equality"})


class UUIDModel(EncryptedModel):
    value = EncryptedUUIDField(queries={"queryType": "equality"})


# Range-queryable models
class BigIntegerModel(EncryptedModel):
    value = EncryptedBigIntegerField(queries={"queryType": "range"})


class DateModel(EncryptedModel):
    value = EncryptedDateField(queries={"queryType": "range"})


class DateTimeModel(EncryptedModel):
    value = EncryptedDateTimeField(queries={"queryType": "range"})


class DecimalModel(EncryptedModel):
    value = EncryptedDecimalField(max_digits=10, decimal_places=2, queries={"queryType": "range"})


class DurationModel(EncryptedModel):
    value = EncryptedDurationField(queries={"queryType": "range"})


class FloatModel(EncryptedModel):
    value = EncryptedFloatField(queries={"queryType": "range"})


class IntegerModel(EncryptedModel):
    value = EncryptedIntegerField(queries={"queryType": "range"})


class PositiveBigIntegerModel(EncryptedModel):
    value = EncryptedPositiveBigIntegerField(queries={"queryType": "range"})


class PositiveIntegerModel(EncryptedModel):
    value = EncryptedPositiveIntegerField(queries={"queryType": "range"})


class PositiveSmallIntegerModel(EncryptedModel):
    value = EncryptedPositiveSmallIntegerField(queries={"queryType": "range"})


class SmallIntegerModel(EncryptedModel):
    value = EncryptedSmallIntegerField(queries={"queryType": "range"})


class TimeModel(EncryptedModel):
    value = EncryptedTimeField(queries={"queryType": "range"})


# An unmanaged model for testing a model that doesn't have its encryption keys
# in the key vault.
class KeyTestModel(models.Model):
    value = EncryptedCharField(max_length=255)

    class Meta:
        managed = False


# Allow querying the key vault using the ORM
class EncryptionKey(models.Model):
    key_alt_name = models.CharField(max_length=500, db_column="keyAltNames")

    class Meta:
        db_table = "__keyVault"
        managed = False
